# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
Parser converting a FZN file to internal model representation.

This parser does not support the complete set of predicates described in the specifications of FlatZinc
that can be found here: http://www.minizinc.org/downloads/doc-1.6/flatzinc-spec.pdf

Basically, it supports essentially integer expressions, some floating point expressions and custom
predicates related to scheduling.

The predicates that are supported are:

 * *array predicates*

   array_bool_and, array_bool_element, array_bool_or, array_bool_xor,
   array_float_element, array_int_element, array_set_element,
   array_var_bool_element, array_var_float_element, array_var_int_element, array_var_set_element.

 * *boolean predicates*

   bool2int, bool_and, bool_clause, bool_eq, bool_eq_reif, bool_le, bool_le_reif,
   bool_lin_eq, bool_lin_le, bool_lt, bool_lt_reif, bool_not, bool_or, bool_xor.

 * *integer predicates*

   int_abs, int_div, int_eq, int_eq_reif, int_le, int_le_reif, int_lin_eq, int_lin_eq_reif,
   int_lin_le, int_lin_le_reif, int_lin_ne, int_lin_ne_reif, int_lt, int_lt_reif, int_max, int_min,
   int_mod, int_ne, int_ne_reif, int_plus, int_times, int2float.

 * *float predicates*

   float_abs, float_exp, float_ln, float_log10, float_log2, float_sqrt, float_eq, float_eq_reif,
   float_le, float_le_reif, float_lin_eq, float_lin_eq_reif, float_lin_le, float_lin_le_reif, float_lin_lt,
   float_lin_lt_reif, float_lin_ne, float_lin_ne_reif, float_lt, float_lt_reif, float_max, float_min,
   float_ne, float_ne_reif, float_plus.

 * *set predicates*

   set_in, set_in_reif.

 * *custom predicates*

   all_different_int, subcircuit, count_eq_const, table_int, inverse,
   lex_lesseq_bool, lex_less_bool, lex_lesseq_int, lex_less_int, int_pow, cumulative


Detailed description
--------------------
"""

from docplex.cp.fzn.fzn_tokenizer import *
from docplex.cp.expression import *
from docplex.cp.solution import *
from docplex.cp.model import CpoModel
import docplex.cp.modeler as modeler
import docplex.cp.config as config
import docplex.cp.expression as expression
import collections
from docplex.cp.utils import xrange
import traceback


###############################################################################
## Constants
###############################################################################

###############################################################################
## Public classes
###############################################################################

class FznParserException(CpoException):
    """ The base class for exceptions raised by the CPO parser
    """
    def __init__(self, msg):
        """ Create a new exception
        Args:
            msg: Error message
        """
        super(FznParserException, self).__init__(msg)

# Parameter descriptor
FznParameter = collections.namedtuple('FznParameter', ('name',   # Variable name
                                                       'type',   # Variable type (string)
                                                       'size',   # Array size (if array), None for single value
                                                       'value',  # Value
                                                       ))

class FznObject(object):
    """ Descriptor of a FZN object
    """
    __slots__ = ()


class FznParameter(FznObject):
    """ Descriptor of a FZN parameter
    """
    __slots__ = ('name',   # Parameter name
                 'type',   # Parameter type
                 'size',   # Array size (if array), None for variable
                 'value',  # Initial value (if any)
                 )
    def __init__(self, name, type, size, value):
        """ Create a new FZN parameter
        Args:
            name:   Name of the parameter
            type:   Type of the parameter
            size:   Array size, None if not array
            value:  Parameter value
        """
        self.name = name
        self.type = type
        self.size = size
        self.value = value

    def __str__(self):
        lstr = [self.name, "(type=", str(self.type)]
        if self.size:
            lstr.append(', size=')
            lstr.append(str(self.size))
            lstr.append(', value=[')
            lstr.append(', '.join(str(x) for x in self.value))
            lstr.append(']')
        else:
            lstr.append(', value=')
            lstr.append(str(self.value))
        lstr.append(')')
        return ''.join(lstr)


class FznVariable(FznObject):
    """ Descriptor of a FZN variable
    """
    __slots__ = ('name',           # Variable name
                 'type',           # Variable type (String)
                 'domain',         # Domain
                 'size',           # Array size (if array), None for variable
                 'value',          # Initial value (if any)
                 'annotations',    # Dictionary of annotations

                 # Attributes needed for model reduction
                 'ref_vars',   # Tuple of variables referenced by this variable
                 )
    def __init__(self, name, type, domain, annotations, size, value):
        """ Create a new FZN variable
        Args:
            name:         Name of the variable
            type:         Variable type
            domain:       Variable domain
            annotations:  Declaration annotations (dictionary)
            size:         Array size, None if not array
            value:        Initial value, None if none
        """
        self.name = name
        self.type = type
        self.domain = domain
        self.annotations = annotations
        self.size = size
        self.value = value

    def is_defined(self):
        """ Check if the variable is introduced
        Return:
            True if variable is introduced, False otherwise
        """
        return 'is_defined_var' in self.annotations

    def is_introduced(self):
        """ Check if the variable is introduced
        Return:
            True if variable is introduced, False otherwise
        """
        return 'var_is_introduced' in self.annotations

    def is_output(self):
        """ Check if the variable is introduced
        Return:
            True if variable is introduced, False otherwise
        """
        return ('output_var' in self.annotations) or ('output_array' in self.annotations)

    def _get_domain_bounds(self):
        """ Get the variable domain bounds
        Return:
            Tuple of values, or single value if identical
        """
        dmin = self.domain[0]
        dmin = dmin[0] if isinstance(dmin, tuple) else dmin
        dmax = self.domain[-1]
        dmax = dmax[-1] if isinstance(dmax, tuple) else dmax
        return (dmin, dmax)

    def __str__(self):
        lstr = [self.name, "(type=", self.type, ", dom=", str(self.domain)]
        if self.is_defined():
            lstr.append(", defined")
        if self.is_introduced():
            lstr.append(", introduced")
        if self.size:
            if self.value:
                lstr.append(', value=[')
                for i, x in enumerate(self.value):
                    if i > 0:
                        lstr.append(', ')
                    if isinstance(x, tuple) and isinstance(x[0], FznVariable):
                        lstr.append("{}[{}]".format(x[0].name, x[1]))
                    elif isinstance(x, FznVariable):
                        lstr.append(x.name)
                    else:
                        lstr.append(str(x))
                lstr.append(']')
            else:
                lstr.append(", size={}".format(self.size))
        elif self.value:
            lstr.append(", value={}".format(self.value))
        lstr.append(')')
        return ''.join(lstr)


class FznConstraint(FznObject):
    """ Descriptor of a FZN constraint
    """
    __slots__ = ('predicate',  # Name of the predicate
                 'args',       # Arguments
                 'defvar',     # Name of the variable defined by this constraint

                 # Attributes needed for model reduction
                 'ref_vars',   # Tuple of variables referenced by this constraint, but not defined
                 )
    def __init__(self, predicate, args, annotations):
        """ Create a new FZN constraint
        Args:
            predicate:    Name of the predicate
            args:         List or arguments
            annotations:  Declaration annotations
        """
        self.predicate = predicate
        self.args = args
        self.defvar = annotations.get('defines_var', (None,))[0]
        self.ref_vars = ()

    def _ref_vars_iterator(self):
        """ Iterator on the variables that are referenced in the arguments of this constraint.
        Returns:
            Iterator on all variables referenced by this constraint
        """
        for a in self.args:
            if is_array(a):
                for v in a:
                    if isinstance(v, FznVariable):
                        yield v
            elif isinstance(a, FznVariable):
                yield a

    def __str__(self):
        lstr = [self.predicate, "("]
        for i, x in enumerate(self.args):
            if i > 0:
                lstr.append(', ')
            if isinstance(x, tuple) and isinstance(x[0], FznVariable):
                lstr.append("{}[{}]".format(x[0].name, x[1]))
            elif isinstance(x, (FznVariable, FznParameter)):
                lstr.append(x.name)
            elif isinstance(x, list):
                lstr.append("[{}]".format(', '.join(str(v) for v in x)))
            else:
                lstr.append(str(x))
        lstr.append(')')
        if self.defvar:
            lstr.append(":")
            lstr.append(self.defvar.name)
        return ''.join(lstr)


class FznObjective(FznObject):
    """ Descriptor of a FZN objective
    """
    __slots__ = ('operation',   # Objective operation in 'satisfy', 'minimize', 'maximize'
                 'expr',        # Target expression
                 'annotations', # Annotations
                )

    def __init__(self, operation, expr, annotations):
        """ Create a new FZN constraint
        Args:
            operation:   Objective operation in 'satisfy', 'minimize', 'maximize'
            expr:        Target expression
            annotations: Annotations
        """
        self.operation   = operation
        self.expr        = expr
        self.annotations = annotations

    def __str__(self):
        return "{} {} ({})".format(self.operation, self.expr, self.annotations)


class FznReader(object):
    """ Reader of FZN file format """
    __slots__ = ('source_file',  # Source file
                 'tokenizer',    # Reading tokenizer
                 'token',        # Last read token
                 'var_map',      # Dictionary of variables.
                                 # Key is variable name, value is variable descriptor
                 'parameters',   # List of parameters
                 'variables',    # List of variables
                 'constraints',  # List of model constraints
                 'objective',    # Model objective
                 )

    def __init__(self, mdl=None):
        """ Create a new FZN reader
        """
        super(FznReader, self).__init__()
        self.source_file = None
        self.tokenizer = None
        self.token = None
        self.var_map = {}
        self.parameters = []
        self.variables = []
        self.constraints = []
        self.objective = None


    def parse(self, cfile):
        """ Parse a FZN file

        Args:
            cfile: FZN file to read
        Raises:
            FznParserException: Parsing exception
        """
        # Store file name if first file
        self.source_file = cfile
        self.tokenizer = FznTokenizer(file=cfile)
        self._read_document()
        self.tokenizer = None


    def parse_string(self, str):
        """ Parse a string

        Result of the parsing is added to the current result model.

        Args:
            str: String to parse
        """
        self.tokenizer = FznTokenizer(input=str)
        self._read_document()
        self.tokenizer = None


    def write(self, out=None):
        """ Write the model.

        If the given output is a string, it is considered as a file name that is opened by this method
        using 'utf-8' encoding.

        Args:
            out (Optional): Target output stream or file name. If not given, default value is sys.stdout.
        """
        # Check file
        if is_string(out):
            with open_utf8(os.path.abspath(out), mode='w') as f:
                self.write(f)
                return
        # Check default output
        if out is None:
            out = sys.stdout

        # Write model content
        for x in self.parameters:
            print(str(x))
        for x in self.variables:
            print(str(x))
        for x in self.constraints:
            print(str(x))
        out.flush()


    def _read_document(self):
        """ Read all FZN document
        """
        try:
            self._next_token()
            while self._read_predicate():
                pass
            while self._read_parameter_or_variable():
                pass
            while self._read_constraint():
                pass
            self._read_objective()
        except Exception as e:
            if isinstance(e, FznParserException):
                raise e
            if config.context.log_exceptions:
                traceback.print_exc()
            self._raise_exception(str(e))

        if self.token is not TOKEN_EOF:
            self._raise_exception("Unexpected token '{}'".format(self.token))


    def _read_predicate(self):
        """ Read a predicate declaration

        This function is called with first token already read and terminates with next token already read.

        Returns:
            True if a predicate has been read, False if nothing to process
        """
        if self.token is not TOKEN_KEYWORD_PREDICATE:
            return False

        # Read predicate declaration
        while self.token not in (TOKEN_SEMICOLON, TOKEN_EOF):
            self._next_token()
        if self.token is not TOKEN_SEMICOLON:
            self._raise_exception("Semicolon ';' expected at the end of a predicate declaration.")
        self._next_token()
        return True


    def _read_parameter_or_variable(self):
        """ Read a parameter or variable declaration

        This function is called with first token already read and terminates with next token already read.

        Returns:
            True if a parameter has been read, False if nothing to process
        """
        tok = self.token
        if tok.type is not TOKEN_TYPE_KEYWORD:
            return False

        # Read array size if any
        arsize = self._read_array_size()

        # Check if variable declaration
        tok = self.token
        if tok is TOKEN_KEYWORD_VAR:
            self._next_token()
            return self._read_variable(arsize)

        # Check type name
        if tok not in (TOKEN_KEYWORD_BOOL, TOKEN_KEYWORD_FLOAT, TOKEN_KEYWORD_INT, TOKEN_KEYWORD_SET):
            return False
        typ = tok
        if typ is TOKEN_KEYWORD_SET:
            self._check_token(self._next_token(), TOKEN_KEYWORD_OF)
            self._check_token(self._next_token(), TOKEN_KEYWORD_INT)

        # Check separating colon
        self._check_token(self._next_token(), TOKEN_COLON)

        # Check parameter name
        tok = self._next_token()
        if tok.type is not TOKEN_TYPE_SYMBOL:
            self._raise_exception("Symbol expected as parameter name.")
        pid = tok.value
        self._check_token(self._next_token(), TOKEN_ASSIGN)

        # Read expression
        self._next_token()
        expr = self._read_expression()
        if arsize:
            expr = list(expression._domain_iterator(expr))
        if typ is TOKEN_KEYWORD_SET:
            arsize = len(expr)

        self._check_token(self.token, TOKEN_SEMICOLON)
        self._next_token()

        # Build result
        fp = FznParameter(pid, typ.value, arsize, expr)
        self.var_map[pid] = fp
        self.parameters.append(fp)

        return True


    def _read_variable(self, arsize):
        """ Read a variable declaration

        This function is called with first token already read and terminates with next token already read.

        Args:
            arsize:  Array size if any
        Returns:
            True if a variable has been read, False if nothing to process
        """
        # Read type and domain
        typ, dom = self._read_var_domain()
        self._check_token(self.token, TOKEN_COLON)
        tok = self._next_token()
        if tok.type is not TOKEN_TYPE_SYMBOL:
            self._raise_exception("Symbol expected as variable name.")
        vid = tok.value
        tok = self._next_token()

        # Check annotations
        annotations = self._read_annotations()
        # print("Annotations: {}".format(annotations))

        # Check expression
        expr = None
        if self.token is TOKEN_ASSIGN:
            self._next_token()
            expr = self._read_expression()

        # Read ending semicolon
        self._check_token(self.token, TOKEN_SEMICOLON)
        self._next_token()

        # Create variable
        fv = FznVariable(vid, typ.value, dom, annotations, arsize, expr)
        self.var_map[vid] = fv
        self.variables.append(fv)

        return True


    def _read_constraint(self):
        """ Read a constraint

        This function is called with first token already read and terminates with next token already read.

        Returns:
            True if a variable has been read, False if nothing to process
        """
        # Check constraint token
        if self.token is not TOKEN_KEYWORD_CONSTRAINT:
            return False

        # Read constraint name
        tok = self._next_token()
        if tok.type is not TOKEN_TYPE_SYMBOL:
            self._raise_exception("Constraint name '{}' should be a symbol.".format(tok))
        cname = tok.value

        # Read parameters
        args = []
        self._check_token(self._next_token(), TOKEN_PARENT_OPEN)
        self._next_token()
        while self.token is not TOKEN_PARENT_CLOSE:
            args.append(self._read_expression())
            if self.token is TOKEN_COMMA:
                self._next_token()
        self._next_token()

        # Check annotations
        annotations = self._read_annotations()
        defvar = annotations.get('defines_var', (None,))[0]

        # Read ending semicolon
        self._check_token(self.token, TOKEN_SEMICOLON)
        self._next_token()

        # Store constraint
        self.constraints.append(FznConstraint(cname, args, annotations))

        return True


    def _read_objective(self):
        """ Read solve objective

        This function is called with first token already read and terminates with next token already read.

        Returns:
            True if a variable has been read, False if nothing to process
        """
        # Check constraint token
        if self.token is not TOKEN_KEYWORD_SOLVE:
            return False
        self._next_token()

        # Check annotations
        annotations = self._read_annotations()

        # Read solve objective
        tok = self.token
        if (tok not in (TOKEN_KEYWORD_SATISFY, TOKEN_KEYWORD_MINIMIZE, TOKEN_KEYWORD_MAXIMIZE)):
            self._raise_exception(
                "Solve objective '{}' should be a symbol in 'satisfy', 'minimize', 'maximize'.".format(tok))
        obj = tok
        self._next_token()

        # Read expression if any
        expr = None if obj is TOKEN_KEYWORD_SATISFY else self._read_expression()

        # Read ending semicolon
        self._check_token(self.token, TOKEN_SEMICOLON)
        self._next_token()

        # Store objective
        self.objective = FznObjective(obj.value, expr, annotations)

        return True


    def _read_expression(self):
        """ Read an expression

        First expression token is already read.
        Function exits with current token following the last expression token

        Returns:
            Expression that has been read
        """
        tok = self.token
        self._next_token()

        # Check int constant
        if tok.type is TOKEN_TYPE_INTEGER:
            v1 = int(tok.value)
            # Check set const
            if self.token is TOKEN_INTERVAL:
                tok2 = self._next_token()
                if tok2.type is not TOKEN_TYPE_INTEGER:
                    self._raise_exception("Set upper bound {} should be an integer constant.".format(tok2))
                self._next_token()
                v2 = int(tok2.value)
                return (v1,) if v1 == v2 else (v1, v2)
            else:
                return v1

        # Check float constant
        if tok.type is TOKEN_TYPE_FLOAT:
            return float(tok.value)

        # Set of integer constant
        if tok is TOKEN_BRACE_OPEN:
            lints = []
            tok = self.token
            while tok is not TOKEN_BRACE_CLOSE:
                if tok.type is not TOKEN_TYPE_INTEGER:
                    self._raise_exception("Set element {} should be an integer constant.".format(tok))
                lints.append(int(tok.value))
                tok = self._next_token()
                if tok is TOKEN_COMMA:
                    tok = self._next_token()
            self._next_token()
            return lints

        # Check symbols
        if tok.type is TOKEN_TYPE_SYMBOL:
            sid = tok.value
            # Check array access
            if self.token is TOKEN_HOOK_OPEN:
                # Access corresponding FZN element
                elem = self.var_map.get(sid)
                if elem is None:
                    self._raise_exception("Unknown symbol '{}'.".format(sid))
                tok2 = self._next_token()
                if tok2.type is not TOKEN_TYPE_INTEGER:
                    self._raise_exception("Array index '{}' should be an integer constant.".format(tok2))
                self._check_token(self._next_token(), TOKEN_HOOK_CLOSE)
                self._next_token()
                # Build array access as a tuple (arr_name, index)
                return (elem, int(tok2.value))
            # Check annotation function call
            elif self.token is TOKEN_PARENT_OPEN:
                lexprs = [sid]
                self._next_token()
                while self.token is not TOKEN_PARENT_CLOSE:
                    lexprs.append(self._read_expression())
                    if self.token is TOKEN_COMMA:
                        self._next_token()
                self._next_token()
                return tuple(lexprs)
            else:
                # Check if corresponds to FZN element
                return self.var_map.get(sid, sid)

        # Array of expressions
        if tok is TOKEN_HOOK_OPEN:
            lexprs = []
            while self.token is not TOKEN_HOOK_CLOSE:
                lexprs.append(self._read_expression())
                if self.token is TOKEN_COMMA:
                    self._next_token()
            self._next_token()
            return lexprs

        # Check boolean constant
        if tok is TOKEN_KEYWORD_TRUE:
            return True
        if tok is TOKEN_KEYWORD_FALSE:
            return False

        # Unknown
        self._raise_exception("Invalid expression start: '{}'.".format(tok))


    def _read_array_size(self):
        """ Read an array size declaration

        First expression token is already read.
        Function exits with current token following the last expression token

        Returns:
            Array size as int if given,
            -1 if size is not precised,
            None if no array specified
        """
        # Check array token
        if self.token is not TOKEN_KEYWORD_ARRAY:
            return None

        # Read array specs
        self._check_token(self._next_token(), TOKEN_HOOK_OPEN)
        tok = self._next_token()
        if tok is TOKEN_KEYWORD_INT:
            arsize = -1
        else:
            if (tok.type is not TOKEN_TYPE_INTEGER) and tok.value != '1' :
                self._raise_exception("Array size should start by  '1'")
            self._check_token(self._next_token(), TOKEN_INTERVAL)
            tok = self._next_token()
            if tok.type is not TOKEN_TYPE_INTEGER:
                self._raise_exception("Array size '{}' should be integer.".format(tok))
            arsize = int(tok.value)
        self._check_token(self._next_token(), TOKEN_HOOK_CLOSE)
        self._check_token(self._next_token(), TOKEN_KEYWORD_OF)
        self._next_token()

        return arsize


    def _read_var_domain(self):
        """ Read the domain of a variable.

        First expression token is already read.
        Function exits with current token following the last expression token

        Returns:
            Type token and variable domain
        """
        # Get token
        tok = self.token
        typ = tok

        # Check boolean domain
        if typ is TOKEN_KEYWORD_BOOL:
            self._next_token()
            return typ, BINARY_DOMAIN

        # Check undefined domain
        if typ is TOKEN_KEYWORD_INT:
            self._next_token()
            return typ, DEFAULT_INTEGER_VARIABLE_DOMAIN

        # Read set of integers or interval
        if tok is TOKEN_BRACE_OPEN:
            lint = sorted(self._read_expression())
            dom = []
            llen = len(lint)
            i = 0
            while i < llen:
                j = i + 1
                while (j < llen) and (lint[j] == lint[j - 1] + 1):
                    j += 1
                if (j > i + 1):
                    dom.append((lint[i], lint[j - 1]))
                else:
                    dom.append(lint[i])
                i = j
            return TOKEN_KEYWORD_INT, tuple(dom)

        # Check integer domain
        if tok.type is not TOKEN_TYPE_INTEGER:
            self._raise_exception("Variable domain should start by an integer constant.")
        self._next_token()
        if self.token is TOKEN_INTERVAL:
            tok2 = self._next_token()
            if tok2.type is not TOKEN_TYPE_INTEGER:
                self._raise_exception("Domain upper bound {} should be an integer constant.".format(tok2))
            self._next_token()
            v1 = int(tok.value)
            v2 = int(tok2.value)
            if v1 == v2:
                return TOKEN_KEYWORD_INT, (v1,)
            return TOKEN_KEYWORD_INT, ((v1, v2),)
        else:
            return TOKEN_KEYWORD_INT, (int(tok.value),)


    def _read_annotations(self):
        """ Read a list of annotations

        First expression token is already read.
        Function exits with current token following the last expression token

        Returns:
            Dictionary of annotations. Key is name, value is tuple or parameters.
        """
        result = {}

        # Check annotation start token
        while self.token is TOKEN_DOUBLECOLON:
            # Read annotation name
            anm = self._next_token()
            if anm.type is not TOKEN_TYPE_SYMBOL:
                self._raise_exception("Annotation name '{}' should be a symbol.".format(anm))
            args = []
            tok = self._next_token()
            if tok is TOKEN_PARENT_OPEN:
                self._next_token()
                while self.token is not TOKEN_PARENT_CLOSE:
                    args.append(self._read_expression())
                    if self.token is TOKEN_COMMA:
                        self._next_token()
                self._next_token()
            result[anm.value] = tuple(args)

        return result


    def _next_token(self):
        """ Read next token
        Returns:
            Next read token, None if end of input
        """
        self.token = self.tokenizer.next_token()
        # print("Line {}, col {}, tok '{}'".format(self.tokenizer.line_number, self.tokenizer.read_index, self.token))
        return self.token


    def _check_token(self, tok, etok):
        """ Check that a read token is a given one an raise an exception if not
        Args:
            tok: Read token
            etok: Expected token
        """
        if tok is not etok:
            self._raise_exception("Read token '{}' instead of expected '{}'".format(tok, etok))


    def _raise_exception(self, msg):
        """ Raise a Parsing exception
        Args:
            msg:  Exception message
        """
        raise FznParserException(self.tokenizer.build_error_string(msg))



class FznParser(object):
    """ Reader of FZN file format """
    __slots__ = ('model',          # Read model
                 'compiled',       # Model compiled indicator
                 'reader',         # FZN reader
                 'cpo_exprs',      # Dictionary of CPO expressions. Key=name, value=CPO expr
                 'reduce',         # Reduce model indicator
                 'interval_gen',   # Name generator for interval var expressions
                 'cumul_gen',      # Name generator for cumul atom expressions

                 'parameters',     # List of parameters
                 'variables',      # List of variables
                 'constraints',    # List of model constraints
                 'objective',      # Model objective

                 'cur_constraint', # Currently compiled constraint descriptor

                 'def_var_exprs',  # List of expressions waiting for defvars to be defined
                 'cpo_variables',  # Set of names of variables that are translated as real CPO variables
                 )

    def __init__(self, mdl=None):
        """ Create a new FZN format parser

        Args:
            mdl:  Model to fill, None (default) to create a new one.
        """
        super(FznParser, self).__init__()
        self.model = mdl if mdl is not None else CpoModel()
        self.compiled = False
        self.reader = FznReader()
        self.interval_gen = IdAllocator("IntervalVar_")
        self.cumul_gen = IdAllocator("VarCumulAtom_")

        # Do not store location information (would store parser instead of real lines)
        self.model.source_loc = False

        # Set model reduction indicator
        self.reduce = config.context.parser.fzn_reduce


    def get_model(self):
        """ Get the model that have been parsed

        Return:
            CpoModel result of the parsing
        """
        if not self.compiled:
            self.compiled = True
            self._compile_to_model()
        return self.model


    def parse(self, cfile):
        """ Parse a FZN file

        Args:
            cfile: FZN file to read
        Raises:
            FznParserException: Parsing exception
        """
        if self.model.source_file is None:
            self.model.source_file = cfile
        self.reader.parse(cfile)


    def parse_string(self, str):
        """ Parse a string

        Result of the parsing is added to the current result model.

        Args:
            str: String to parse
        """
        self.reader.parse_string(str)


    def get_output_variables(self):
        """ Get the list of model output variables

        Returns:
            List of output variables, in declaration order.
        """
        return [v for v in self.variables if v.is_output()]


    def _write_model(self, out=None):
        """ Print read model (short version)
        Args:
            out (optional): Output stream. Default is stdout
        """
        if out is None:
            out = sys.stdout
        out.write(self.get_model().get_cpo_string(short_output=True))
        out.write("\n")


    def _get_cpo_expr_map(self):
        """ For testing, get the map of CPO expressions
        """
        self.get_model()
        return self.cpo_exprs


    def _compile_to_model(self):
        """ Compile FZN model into CPO model
        """
        # Initialize processing
        self.cpo_exprs = {}
        self.parameters  = self.reader.parameters
        self.variables   = self.reader.variables
        self.constraints = self.reader.constraints
        self.objective   = self.reader.objective
        self.def_var_exprs = {}
        self.cpo_variables = set()

        # Reduce model if required
        if self.reduce:
            self._reduce_model()

        # print("=== Variables:")
        # for v in self.variables:
        #     print(" : {}".format(v))
        # print("=== Constraints:")
        # for c in self.constraints:
        #     print(" : {}".format(c))
        # sys.stdout.flush()

        # Compile parameters
        for x in self.parameters:
            self._compile_parameter(x)

        # Compile variables
        for x in self.variables:
            self._compile_variable(x)

        # Compile constraints
        if self.reduce:
            for x in self.constraints:
                if isinstance(x, FznVariable):
                    self._compile_variable(x)
                else:
                    self._compile_constraint(x)
        else:
            for x in self.constraints:
                self._compile_constraint(x)

        # Compile objective
        self._compile_objective(self.objective)


    def _compile_parameter(self, fp):
        """ Compile a FZN parameter into CPO model
        Args:
            fp: Flatzinc parameter, object of class FznParameter
        """
        if fp.type in ('int', 'bool'):
            expr = CpoValue(fp.value, Type_IntArray if fp.size else Type_Int)
        elif fp.type == 'float':
            expr = CpoValue(fp.value, Type_FloatArray if fp.size else Type_Float)
        else:
            expr = build_cpo_expr(fp.value)

        # Add to map
        expr.set_name(fp.name)
        self.cpo_exprs[fp.name] = expr


    def _compile_variable(self, fv):
        """ Compile a FZN variable into CPO model
        Args:
            fv: Flatzinc variable
        """
        # Check if variable is array
        val = fv.value
        if fv.size:
            # Build array of variables
            if val:
                # Check if there is a reference to a not yet defined variable
                if self.reduce:
                    for v in val:
                        if v in self.def_var_exprs:
                            self.def_var_exprs[v].append(fv)
                            return
                arr = [self._get_cpo_expr(e) for e in val]
                expr = CpoValue(arr, Type_IntVarArray if all(x.type == Type_IntVar for x in arr) else Type_IntExprArray)
            else:
                # Build array of variables
                arr = [integer_var(name=fv.name + '[' + str(i + 1) + ']', domain=fv.domain) for i in range(fv.size)]
                expr = CpoValue(arr, Type_IntVarArray)
        else:
            # Build single variable
            if self.reduce and val:
                if is_int(val):
                    expr = CpoValue(val, Type_Int)
                elif is_bool(val):
                    expr = CpoValue(val, Type_Bool)
                else:
                    expr = self._get_cpo_expr(val)
            else:
                # Check if value is another variable
                if isinstance(val, FznVariable):
                    # Retrieve existing variable
                    expr = self.cpo_exprs.get(val.name)
                    assert isinstance(expr, CpoIntVar), "Variable '{}' not found".format(val.name)
                else:
                    # Create new variable
                    dom = _build_domain(val) if val else fv.domain
                    expr = integer_var(domain=dom)
        expr.set_name(fv.name)
        self.cpo_exprs[fv.name] = expr


    def _compile_constraint(self, fc):
        """ Compile a FZN constraint into CPO model
        Args:
            fv: Flatzinc constraint
        """
        # Search in local methods
        cmeth = getattr(self, "_compile_pred_" + fc.predicate, None)
        if not cmeth:
            raise FznParserException("Predicate '{}' is not supported.".format(fc.predicate))

        # Call compile method
        cmeth(fc)


    def _compile_objective(self, fo):
        """ Compile a FZN objective into CPO model
        Args:
            fo: Flatzinc objective
        """
        #print("Compile objective {}".format(fo))
        if fo is None:
            return
        if fo.operation != 'satisfy':
            expr = self._get_cpo_expr(fo.expr)
            oxpr = modeler.maximize(expr) if fo.operation == 'maximize' else modeler.minimize(expr)
            self._add_to_model(oxpr)


    def _reduce_model(self):
        """ Reduce model size by factorizing expressions when possible
        """
        # Access main model elements
        variables = self.variables
        constraints = self.constraints

        # Build reduction data related to variables
        for fv in variables:
            # Build list of variables that are referenced by this one
            fv.ref_vars = tuple(v for v in fv.value if isinstance(v, FznVariable)) if fv.size else ()
            # Set in defined variables if output
            if fv.is_output():
                self.cpo_variables.add(fv.name)

        # In constraints, replace reference to arrays by arrays themselves
        for fc in constraints:
            fc.args = tuple(a.value if isinstance(a, FznVariable) and a.size else a for a in fc.args)

        # Initialize set of variables defined in constraints
        def_var_map = {}      # Key is variable, value is constraint where variable is defined
        for fc in constraints:
            # print("Scan constraint {}".format(fc))
            defvar = fc.defvar
            if defvar is not None:
                def_var_map[defvar] = fc
            # Build list of all variables referenced by this constraint
            res = set()
            nbrefdvar = 0
            for v in fc._ref_vars_iterator():
                if v is defvar:
                    nbrefdvar += 1
                else:
                    res.add(v)
            fc.ref_vars = tuple(res)
            # Special case for cumulative. All variables are supposed defined by the constraint.
            if fc.predicate == 'cumulative':
                for v in fc.ref_vars:
                    def_var_map[v] = fc
            # print("   result list of ref variables: {}".format(fc.ref_vars))
            # If defined variable is referenced twice in the constraint, remove it as defined (keep as declared variable)
            if nbrefdvar > 1:
                fc.defvar = None
                self.cpo_variables.add(defvar.name)

        # Determine connected variable subsets
        #self._determine_connex_variables(def_var_map)

        # Scan variables to move them after definition of their dependencies when needed
        variables = []  # New list of variables
        for fv in self.variables:
            #print("Scan variable {}, Refvars: {}".format(fv, [v.name for v in fv.ref_vars]))
            if any(v in def_var_map for v in fv.ref_vars):
                if self._insert_in_constraints(fv, 0, def_var_map):
                    # Remove from list of variables (moved in constraints)
                    pass
                else:
                    # Keep it as a model variable
                    self.cpo_variables.add(fv.name)
                    variables.append(fv)
            else:
                variables.append(fv)
        self.variables = variables

        # Reorder constraints
        #print("\nScan constraints. Defined vars: {}".format([v.name for v in defined_vars]))
        constraints = self.constraints
        nbct = len(constraints)
        movedcstr = set() # Constraints already moved
        cx = 0
        while cx < nbct:
            fc = constraints[cx]
            #print("Scan constraint {}. Refvars: {}".format(fc, [v.name for v in fc.ref_vars]))
            #print("Defined vars: {}".format([v.name for v in defined_vars]))
            # Process case of variable that has been inserted in constraints
            if isinstance(fc, FznVariable):
                def_var_map.pop(fc, None)
            else:
                # Search if constraint can be moved
                if (fc not in movedcstr) and (fc.predicate != "cumulative") \
                        and any(v in def_var_map for v in fc.ref_vars) \
                        and self._insert_in_constraints(fc, cx + 1, def_var_map):
                    # Move constraint after all is defined
                    del constraints[cx]
                    movedcstr.add(fc)
                    cx -= 1
                else:
                    # Constraint stays where it is
                    if fc.defvar:
                        def_var_map.pop(fc.defvar, None)
            cx += 1

        # print("Reduction ended.")
        # print("   Variables:")
        # for v in self.variables:
        #     print("      {}".format(v))
        # print("   Constraints:")
        # for c in self.constraints:
        #     print("      {}".format(c))


    def _insert_in_constraints(self, fc, cx, varsctsr):
        """ Insert a constraint or a variable in constraints after all its members are defined
        Args:
            fc:         FZN constraint or variable to insert
            cx:         Start insertion index
            varsctsr:   Map of constraints where each variable is defined
        Return:
            True if insertion was successful, False otherwise
        """
        # Build set of constraints to skip to wait for variable definition
        cset = set(varsctsr[v] for v in fc.ref_vars if v in varsctsr)

        # Check no dependency
        if not cset:
            return False

        # Search in next constraints
        constraints = self.constraints
        for ix in xrange(cx, len(constraints)):
            cset.discard(constraints[ix])
            if not cset:
                # Insert after this place
                constraints.insert(ix + 1, fc)
                # print("   inserted at best rank {}".format(cx + 1))
                return True

        # Impossible to insert
        # print("   impossible to insert")
        return False


    def _determine_connex_variables(self, def_var_map):
        """ Search connex sub-graphes in variables dependencies and put all multi-variables connex parts as cpo variables
        Args:
            def_var_map: Map of constraints where variable is defined
        """

        # Create result map. Key is variable descriptor, value is set of connected variable descriptors
        lcomps = {}

        # For each variable, build the list of depending variables
        for fv in self.variables:
            stack = [fv]
            dset = set(stack)
            while stack:
                sv = stack.pop(0)
                for v in sv.ref_vars:
                    if v not in dset:
                        if v in lcomps:
                            dset.update(lcomps[v])
                        else:
                            dset.add(v)
                            stack.append(v)
                fc = def_var_map.get(fv)
                if fc:
                    for v in fc.ref_vars:
                        if v not in dset:
                            #print("   cstr.ref: {}".format(v))
                            if v in lcomps:
                                dset.update(lcomps[v])
                            else:
                                dset.add(v)
                                stack.append(v)
            lcomps[fv] = dset

        # Sort all variables list in ascending cardinality order
        #print("Variables graphs:")
        #for k, v in lcomps.items():
        #    print("   {}: {}".format(k.name, [x.name for x in v]))
        lcomps = sorted(lcomps.values(), key=lambda x: len(x))

        # Remove each set in next ones
        nbcomps = len(lcomps)
        for x1 in range(nbcomps):
            s1 = lcomps[x1]
            if s1:
                for x2 in range(x1 + 1, nbcomps):
                    lcomps[x2] = lcomps[x2].difference(s1)

        # Set as model variables all that are in component with at least two variables
        for c in lcomps:
            #print("   graph component: {}".format([x.name for x in c]))
            if len(c) > 1:
                for v in c:
                    self.cpo_variables.add(v.name)

        #print("CPO model variables identified: {}".format([v for v in self.cpo_variables]))


    def _get_cpo_expr(self, expr):
        """ Retrieve a CPO expression from its FZN representation
        Args:
            expr:  FZN expression
        Returns:
            Corresponding CPO expression
        """

        #print("   _get_cpo_expr({}, type={})".format(expr, type(expr)))

        # Check basic types
        etyp = type(expr)
        if etyp in (FznVariable, FznParameter):
            v = self.cpo_exprs.get(expr.name)
            if v is None:
                raise FznParserException("Can not find element {}".format(expr.name))
            return v

        # Integer constant
        if etyp in INTEGER_TYPES:
            return CpoValue(expr, Type_Int)

        # Array
        if etyp is list:
            return build_cpo_expr([self._get_cpo_expr(x) for x in expr])

        # Check array access
        if etyp is tuple:
            # Check tuple of integers
            if is_int(expr[0]):
                if len(expr) > 1:
                    return [i for i in range(expr[0], expr[1] + 1)]
                return [expr[0]]
            # Access to array element
            arr = self.cpo_exprs.get(expr[0].name)
            if arr is None:
                raise FznParserException("Can not find array {}".format(expr[0]))
            return(arr.value[expr[1]-1])

        # Boolean
        if etyp in BOOL_TYPES:
            return CpoValue(1, Type_Bool) if expr else CpoValue(0, Type_Bool)

        # String
        if etyp in STRING_TYPES:
            v = self.cpo_exprs.get(expr)
            if v is None:
                raise FznParserException("Can not find element {}".format(expr))
            return v

        # Unknown
        raise FznParserException("Can not find element {}".format(expr))

    # Array predicates

    def _compile_pred_array_bool_element(self, fc):
        self._compile_array_xxx_element(fc)

    def _compile_pred_array_int_element(self, fc):
        self._compile_array_xxx_element(fc)

    def _compile_pred_array_float_element(self, fc):
        self._compile_array_xxx_element(fc)

    def _compile_pred_array_var_bool_element(self, fc):
        self._compile_array_xxx_element(fc)

    def _compile_pred_array_var_int_element(self, fc):
        self._compile_array_xxx_element(fc)

    def _compile_pred_array_var_float_element(self, fc):
        self._compile_array_xxx_element(fc)

    def _compile_pred_array_bool_and(self, fc):
        self._compile_op_assign_arg_1(fc, modeler.min_of)

    def _compile_pred_array_bool_or(self, fc):
        self._compile_op_assign_arg_1(fc, modeler.max_of)

    # Bool predicates

    def _compile_pred_bool_and(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.logical_and)

    def _compile_pred_bool_or(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.logical_or)

    def _compile_pred_bool_xor(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.diff)

    def _compile_pred_bool_not(self, fc):
        self._compile_op_assign_arg_1(fc, modeler.logical_not)

    def _compile_pred_bool_eq(self, fc):
        self._compile_xxx_eq(fc)

    def _compile_pred_bool_eq_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.equal)

    def _compile_pred_bool_le(self, fc):
        self._compile_op_arg_2(fc, modeler.less_or_equal)

    def _compile_pred_bool_le_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.less_or_equal)

    def _compile_pred_bool_lin_eq(self, fc):
        self._compile_scal_prod(fc, modeler.equal)

    def _compile_pred_bool_lin_le(self, fc):
        self._compile_scal_prod(fc, modeler.less_or_equal)

    def _compile_pred_bool_lt(self, fc):
        self._compile_op_arg_2(fc, modeler.less)

    def _compile_pred_bool_lt_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.less)

    def _compile_pred_bool2int(self, fc):
        self._compile_xxx_eq(fc)

    # Float predicates

    def _compile_pred_float_abs(self, fc):
        self._compile_op_assign_arg_1(fc, modeler.abs_of)

    def _compile_pred_float_exp(self, fc):
        self._compile_op_assign_arg_1(fc, modeler.exponent)

    def _compile_pred_float_ln(self, fc):
        self._compile_op_assign_arg_1(fc, modeler.log)

    def _compile_pred_float_log10(self, fc):
        a, r = fc.args
        self._make_equal(r, modeler.log(self._get_cpo_expr(a)) / modeler.log(10), fc.defvar)

    def _compile_pred_float_log2(self, fc):
        a, r = fc.args
        self._make_equal(r, modeler.log(self._get_cpo_expr(a)) / modeler.log(2), fc.defvar)

    def _compile_pred_float_sqrt(self, fc):
        a, r = fc.args
        self._make_equal(r, modeler.power(self._get_cpo_expr(a), 0.5), fc.defvar)

    def _compile_pred_float_eq(self, fc):
        self._compile_xxx_eq(fc)

    def _compile_pred_float_eq_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.equal)

    def _compile_pred_float_le(self, fc):
        self._compile_op_arg_2(fc, modeler.less_or_equal)

    def _compile_pred_float_le_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.less_or_equal)

    def _compile_pred_float_lin_eq(self, fc):
        self._compile_scal_prod(fc, modeler.equal)

    def _compile_pred_float_lin_le(self, fc):
        self._compile_scal_prod(fc, modeler.less_or_equal)

    def _compile_pred_float_lin_lt(self, fc):
        self._compile_scal_prod(fc, modeler.lesst)

    def _compile_pred_float_lin_ne(self, fc):
        self._compile_scal_prod(fc, modeler.diff)

    def _compile_pred_float_lin_eq_reif(self, fc):
        self._compile_scal_prod(fc, modeler.equal, True)

    def _compile_pred_float_lin_le_reif(self, fc):
        self._compile_scal_prod(fc, modeler.less_or_equal, True)

    def _compile_pred_float_lin_lt_reif(self, fc):
        self._compile_scal_prod(fc, modeler.less, True)

    def _compile_pred_float_lin_ne_reif(self, fc):
        self._compile_scal_prod(fc, modeler.diff, True)

    def _compile_pred_float_lt(self, fc):
        self._compile_op_arg_2(fc, modeler.less)

    def _compile_pred_float_lt_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.less)

    def _compile_pred_float_max(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.max_of)

    def _compile_pred_float_min(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.min_of)

    def _compile_pred_float_ne(self, fc):
        self._compile_op_arg_2(fc, modeler.diff)

    def _compile_pred_float_ne_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.diff)

    def _compile_pred_float_plus(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.plus)

    # Int predicates

    def _compile_pred_int_abs(self, fc):
        self._compile_op_assign_arg_1(fc, modeler.abs_of)

    def _compile_pred_int_div(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.int_div)

    def _compile_pred_int_eq(self, fc):
        self._compile_xxx_eq(fc)

    def _compile_pred_int_eq_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.equal)

    def _compile_pred_int_le(self, fc):
        self._compile_op_arg_2(fc, modeler.less_or_equal)

    def _compile_pred_int_le_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.less_or_equal)

    def _compile_pred_int_lin_eq(self, fc):
        self._compile_scal_prod(fc, modeler.equal)

    def _compile_pred_int_lin_le(self, fc):
        self._compile_scal_prod(fc, modeler.less_or_equal)

    def _compile_pred_int_lin_ne(self, fc):
        self._compile_scal_prod(fc, modeler.diff)

    def _compile_pred_int_lin_eq_reif(self, fc):
        self._compile_scal_prod(fc, modeler.equal, True)

    def _compile_pred_int_lin_le_reif(self, fc):
        self._compile_scal_prod(fc, modeler.less_or_equal, True)

    def _compile_pred_int_lin_ne_reif(self, fc):
        self._compile_scal_prod(fc, modeler.diff, True)

    def _compile_pred_int_lt(self, fc):
        self._compile_op_arg_2(fc, modeler.less)

    def _compile_pred_int_lt_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.less)

    def _compile_pred_int_max(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.max_of)

    def _compile_pred_int_min(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.min_of)

    def _compile_pred_int_mod(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.mod)

    def _compile_pred_int_ne(self, fc):
        self._compile_op_arg_2(fc, modeler.diff)

    def _compile_pred_int_ne_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.diff)

    def _compile_pred_int_plus(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.plus)

    def _compile_pred_int_times(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.times)

    def _compile_pred_int2float(self, fc):
        self._compile_xxx_eq(fc)


    # Set predicates

    def _compile_pred_set_in(self, fc):
        self._compile_op_arg_2(fc, modeler.allowed_assignments)

    def _compile_pred_set_in_reif(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.allowed_assignments)


    # Custom predicates

    def _compile_pred_all_different_int(self, fc):
        self._compile_op_arg_1(fc, modeler.all_diff)

    def _compile_pred_count_eq_const(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.count)

    def _compile_pred_lex_lesseq_int(self, fc):
        self._compile_op_arg_2(fc, modeler.lexicographic)

    def _compile_pred_lex_lesseq_bool(self, fc):
        self._compile_op_arg_2(fc, modeler.lexicographic)

    def _compile_pred_int_pow(self, fc):
        self._compile_op_assign_arg_2(fc, modeler.power)

    def _compile_pred_cumulative(self, fc):
        """ Requires that a set of tasks given by start times s, durations d, and resource requirements r,
        never require more than a global resource bound b at any one time.
        Args:
            fc:  Constraint descriptor, with arguments
                stime:  Tasks start time
                tdur:   Tasks tasks durations
                rreq:   Task resource requirements
                bnd:    Global resource bound
        """
        #print("Process cumulative constraint {}".format(fc))

        # Access constraint arguments
        stime, tdur, rreq, bnd = fc.args

        # Create interval vars and cumul atoms
        ls = _get_fzn_array(stime)
        ld = _get_fzn_array(tdur)
        lr = _get_fzn_array(rreq)
        #print("ls: {}\nld: {}\nlr: {}".format(ls, [s for s in ld], lr))
        cumul_atoms = []
        for s, d, r in zip(ls, ld, lr):
            vname = None
            # Get start time
            if isinstance(s, FznVariable):
                ds = s._get_domain_bounds()
                vname = s.name
            else:
                ds = (s, s)
            # Get duration
            if isinstance(d, FznVariable):
                dd = d._get_domain_bounds()
                if vname is None:
                    vname = d.name
            else:
                dd = (d, d)
            # Get requirement
            if isinstance(r, FznVariable):
                dr = r._get_domain_bounds()
                if vname is None:
                    vname = r.name
            else:
                dr = (r, r)
            #print("ds: {}, dd: {}, dr: {}".format(ds, dd, dr))

            # Create interval variable
            if vname is None:
                vname = self.interval_gen.allocate()
            else:
                vname = "Itv_" + vname
                if vname in self.cpo_exprs:
                    cnt = 1
                    nname = vname + "_1"
                    while nname in self.cpo_exprs:
                        cnt += 1
                        nname = vname + "_" + str(cnt)
                    vname = nname
            # Create interval variable
            ivar = self.cpo_exprs.get(vname)
            if ivar:
                # Check it is the same
                assert isinstance(ivar, CpoIntervalVar) and ivar.get_start() == ds and ivar.get_size() == dd and ivar.get_end() == (INTERVAL_MIN, INTERVAL_MAX)
            else:
                ivar = interval_var(start=ds, end=(INTERVAL_MIN, INTERVAL_MAX), size=dd, name=vname)
                self.cpo_exprs[vname] = ivar

            # Create pulse
            pulse = modeler.pulse(ivar, dr)
            cumul_atoms.append(pulse)
            # Replace previous variable by access to interval variable
            if isinstance(s, FznVariable):
                self._assign_to_var(s, modeler.start_of(ivar))
            if isinstance(d, FznVariable):
                self._assign_to_var(d, modeler.size_of(ivar))
            if isinstance(r, FznVariable):
                self._assign_to_var(r, modeler.height_at_start(ivar, pulse))

        # Create final constraint
        cumf = CpoFunctionCall(Oper_sum, Type_CumulFunction, (CpoValue(cumul_atoms, Type_CumulAtomArray),))
        self._add_to_model(modeler.greater_or_equal(self._get_cpo_expr(bnd), cumf))


    def _compile_pred_subcircuit(self, fc):
        """ Constrains the elements of x to define a subcircuit where x[i] = j means that j is the successor of i and x[i] = i means that i is not in the circuit.
        """
        x = [0] + list(self._get_cpo_expr(fc.args[0]).value)
        self._add_to_model(CpoFunctionCall(Oper__sub_circuit, Type_Constraint, (build_cpo_expr(x),) ))


    def _compile_pred_inverse(self, fc):
        """ Constrains two arrays of int variables, f and invf, to represent inverse functions. All the values in each array must be within the index set of the other array.
        Args:
            f:     First function as array of int
            invf:  Inverse function
        """
        f, invf = fc.args
        f = [0] + list(self._get_cpo_expr(f).value)
        invf = [0] + list(self._get_cpo_expr(invf).value)
        self._add_to_model(CpoFunctionCall(Oper_inverse, Type_Constraint, (build_cpo_expr(f), build_cpo_expr(invf))))


    def _compile_pred_bool_clause(self, fc):
        """ Implementation of bool_clause predicate
        Args:
            a:  First array of booleans
            b:  Second array of booleans
        """
        a, b = fc.args
        # Default implementation
        exprs = list(self._get_cpo_expr(a).value)
        for x in self._get_cpo_expr(b).value:
            exprs.append(1 - x)
        self._add_to_model(modeler.sum_of(exprs) >= 1)

        # Alternative implementation
        # self._add_to_model( (modeler.max_of(a) > 0) | (modeler.min_of(b) == 0) )

        # Other alternative implementation
        # expr = None
        # for x in _get_value(a):
        #     x = x > 0
        #     expr = x if expr is None else modeler.logical_or(expr, x)
        # for x in _get_value(b):
        #     x = x == 0
        #     expr = x if expr is None else modeler.logical_or(expr, x)
        # self._add_to_model(expr)


    def _compile_pred_table_int(self, fc):
        """ Implement custom predicate table_int
        Args:
            vars:    Array of variables
            values:  List of values
        """
        vars, values = fc.args
        # Split value array in tuples
        vars = self._get_cpo_expr(vars).value
        tsize = len(vars)
        if tsize != 0:
            values = self._get_cpo_expr(values).value
            tuples = [values[i: i + tsize] for i in range(0, len(values), tsize)]
            # Build allowed assignment expression
            self._add_to_model(modeler.allowed_assignments(vars, tuples))


    def _compile_pred_lex_less_bool(self, fc):
        """ Requires that the array vars1 is strictly lexicographically less than array vars2
        Args:
            vars1:  First array of variables
            vars2:  Second array of variables
        """
        vars1, vars2 = fc.args
        # Add 0 and 1 at the end of arrays to force inequality
        vars1 = list(self._get_cpo_expr(vars1).value) + [1]
        vars2 = list(self._get_cpo_expr(vars2).value) + [0]
        self._add_to_model(modeler.lexicographic(vars1, vars2))


    def _compile_pred_lex_less_int(self, fc):
        self._compile_pred_lex_less_bool(fc)


    def _get_domain_bounds(self, x):
        """ Get min and max bounds of an expression, integer variable or integer
        Args:
            expr: CPO integer variable or expression
        Returns:
            Tuple (min, max)
        """
        # Case of variable or variable replaced by an expression
        if isinstance(x, CpoIntVar):
            return (x.get_domain_min(), x.get_domain_max())
        if isinstance(x, CpoFunctionCall):
            if x.operation is Oper_start_of:
                return x.children[0].get_start()
            if x.operation is Oper_size_of:
                return x.children[0].get_size()
            # if x.is_kind_of(Type_BoolExpr):
            #     return (0, 1)
            cpov = self.cpo_exprs.get(x.name)
            if isinstance(cpov, CpoIntVar):
                return (cpov.get_domain_min(), cpov.get_domain_max())
            raise FznParserException("Unknow expression to take bounds from: {}".format(x))
        if isinstance(x, CpoValue):
            x = x.value
        if is_int(x):
            return (x, x)

        return x


    def _assign_to_var(self, var, expr):
        """ Set a identifier with an expression
        Args:
            var:  Target FZN variable
            expr: CPO expression to assign
        Returns:
            None. If needed, changes are done in reader context.
        """
        #print("_assign_to_var {} expression {}".format(var, expr))

        # Retrieve existing expression
        vname = var.name
        vexpr = self.cpo_exprs.get(vname)

        # Check if reduction
        if (not self.reduce) or (vname in self.cpo_variables) or (vexpr is not None and not isinstance(vexpr, CpoIntVar)):
            self._add_to_model(modeler.equal(vexpr, expr))
        else:
            # Assign new name to expression
            self.cpo_exprs[vname] = expr
            expr.set_name(vname)
            # Constrain expression to variable domain
            self._constrain_expr_domain(expr, var)


    def _make_equal(self, var, expr, dvar):
        """ Make equal two FZN expressions
        Args:
            var:   FZN variable or value
            expr:  CPO expression to be equal with
            dvar:  Define variable, None if none
        Returns:
            None. If needed, changes are done in reader context.
        """

        # Check if var is not a variable
        if not isinstance(var, FznVariable):
            self._add_to_model(modeler.equal(expr, self._get_cpo_expr(var)))
            return

        # Check no reduction
        if not self.reduce or (dvar is not var):
            self._add_to_model(modeler.equal(self._get_cpo_expr(var), expr))
            return

        # Assign to variable
        self._assign_to_var(var, expr)


    def _constrain_expr_domain(self, expr, var):
        """ Constrain the domain of an expression to the domain of a variable
        Args:
            expr: CPO expression to constrain
            var:  CPO or FZN integer var to take domain from
        """
        dom = var.domain
        #print("   constrain expression {} to domain {}".format(expr, dom))
        dmin = expression._domain_min(dom)
        dmax = expression._domain_max(dom)

        # Check boolean expression
        if dmin == 0 and dmax == 1:
            return
        if expr.is_kind_of(Type_BoolExpr) and (dmin <= 0) and (dmax >= 1):
            return

        # Add appropriate constraint
        if len(dom) == 1:  # Single segment
            # Use range
            if dmin == dmax:
                self._add_to_model(modeler.equal(expr, dmin))
            else:
                self._add_to_model(modeler.range(expr, dmin, dmax))
        else:
            # Use allowed assignment
            self._add_to_model(modeler.allowed_assignments(expr, expression._domain_iterator(dom)))


    def _compile_op_assign_arg_1(self, fc, op):
        """ Compile operation with single argument equal to a result
        Args:
            fc:  Constraint descriptor
            op:  CPO operation to apply to first argument
        """
        # Access constraint arguments
        a, r = fc.args
        # Assign to expression
        self._make_equal(r, op(self._get_cpo_expr(a)), fc.defvar)


    def _compile_op_assign_arg_2(self, fc, op):
        """ Compile operation with two arguments equal to a result
        Args:
            fc:  Constraint descriptor
            op:  CPO operation to apply to arguments
        """
        # Access constraint arguments
        a, b, r = fc.args
        # Assign to expression
        self._make_equal(r, op(self._get_cpo_expr(a), self._get_cpo_expr(b)), fc.defvar)


    def _compile_op_arg_1(self, fc, op):
        """ Compile operation with one arguments and no result
        Args:
            fc:  Constraint descriptor
            op:  CPO operation to apply to argument
        """
        self._add_to_model(op(self._get_cpo_expr(fc.args[0])))


    def _compile_op_arg_2(self, fc, op):
        """ Compile operation with two arguments and no result
        Args:
            fc:  Constraint descriptor
            op:  CPO operation to apply to arguments
        """
        a, b = fc.args
        self._add_to_model(op(self._get_cpo_expr(a), self._get_cpo_expr(b)))


    def _compile_array_xxx_element(self, fc):
        """ Compile access to array element """
        x, t, r = fc.args
        if not is_int(x):
            x = self._get_cpo_expr(x)
        self._make_equal(r, modeler.element(self._get_cpo_expr(t), x - 1), fc.defvar)


    def _compile_xxx_eq(self, fc):
        """ Compile all equality predicates """

        # Access constraint arguments
        a, b = fc.args

        # Check default case
        if not self.reduce:
            self._add_to_model(modeler.equal(self._get_cpo_expr(a), self._get_cpo_expr(b)))
            return

        # Process trivial cases
        if a is b:
            return None

        # Retrieve defined variable
        defvar = fc.defvar
        if defvar is None:
            self._add_to_model(modeler.equal(self._get_cpo_expr(a), self._get_cpo_expr(b)))
            return

        if defvar is a:
            self._assign_to_var(a, self._get_cpo_expr(b))
        else:
            self._assign_to_var(b, self._get_cpo_expr(a))


    def _compile_scal_prod(self, fc, op, reif=False):
        """ Compile a scalar product
        Args:
            fc:  Constraint
            op:  Comparison operation
        """
        # Access constraint arguments
        if reif:
            coefs, vars, res, reif = fc.args
        else:
            coefs, vars, res = fc.args

        # Check no reduction
        defvar = fc.defvar
        if not self.reduce or not defvar:
            expr = op(modeler.scal_prod(self._get_cpo_expr(coefs), self._get_cpo_expr(vars)), self._get_cpo_expr(res))
            if reif:
                expr = modeler.equal(self._get_cpo_expr(reif), expr)
            self._add_to_model(expr)
            return

        # Get array elements
        coefs = _get_fzn_array(coefs)
        vars = _get_fzn_array(vars)

        # Check if defined variable is the result
        if defvar is res or defvar is reif:
            expr = self._build_scal_prod_expr(coefs, vars, 0)
        else:
            # Arrange expression to have defined variable on the left
            vx = vars.index(defvar)
            vcoef = coefs[vx]
            vars = vars[:vx] + vars[vx + 1:]
            coefs = coefs[:vx] + coefs[vx + 1:]
            expr = res if is_int(res) else self._get_cpo_expr(res)
            if vcoef < 0:
                vcoef = -vcoef
                expr = -expr
            else:
                coefs = list([-c for c in coefs])

            # Build result
            expr = self._build_scal_prod_expr(coefs, vars, expr)
            if vcoef != 1:
                expr = expr / vcoef

        # Check reif
        if reif:
            expr = op(expr, self._get_cpo_expr(res))
            self._assign_to_var(defvar, expr)
        else:
            # Check equality with a variable
            if op is modeler.equal:
                self._assign_to_var(defvar, expr)
            else:
                self._add_to_model(op(expr, self._get_cpo_expr(defvar)))


    def _build_scal_prod_expr(self, coefs, vars, res):
        """ Build a scal prod expression
        Args:
            coefs:  Array of coefficients (integers)
            vars:   Array of FZN variables
            res:    Initial result value (integer)
        Returns:
            New CPO scal_prod expression
        """
        # Build array of CPO variables
        vars = [self._get_cpo_expr(v) for v in vars]

        # Check developed scal_prod
        if len(coefs) <= 2 or (all(c == 1 or c == -1 for c in coefs)):
            for c, v in zip(coefs, vars):
                if c != 0:
                    if res is 0:
                        res = _mutl_by_int(v, c)
                    elif c < 0:
                        res = res - _mutl_by_int(v, -c)
                    else:
                        res = res + _mutl_by_int(v, c)
            return res

        # Build normal scal_prod
        expr = modeler.scal_prod(coefs, vars)
        if res is not 0:
            expr = res + expr
        return expr


    def _add_to_model(self, expr):
        """ Add an expression to the CPO model
        Args:
            expr: CPO expression to add
        """
        #print("_add_to_model({})".format(expr))
        self.model.add(expr)
        
        # Scan expression to identify used variables
        estack = [expr]
        doneset = set()  # Set of expressions already processed
        while estack:
            e = estack.pop()
            eid = id(e)
            if not eid in doneset:
                doneset.add(eid)
                if e.type.is_variable:
                    #print("   add CPO variable {}".format(e))
                    self.cpo_variables.add(e.name)
                # Stack children expressions
                estack.extend(e.children)


    def _write(self, out=None):
        """ Write current parser status
        Args:
            out (optional):  Write output. sys.stdout if not given
        """
        if out is None:
            out = sys.stdout
        out.write("Reader status:\n")
        out.write("   CPO expressions:\n")
        for k in sorted(self.cpo_exprs.keys()):
            v = self.cpo_exprs[k]
            out.write("      {}: {} ({})\n".format(k, v, type(v)))
        out.write("   Model expressions:\n")
        for x in self.model.get_all_expressions():
            out.write("      {}\n".format(x[0]))


###############################################################################
## Utility functions
###############################################################################

def _get_fzn_array(fzo):
    """ Get the list of objects of an FZN object
    Args:
        fzo: FZN object
    Returns:
        Array of FZN objects
    """
    return fzo if isinstance(fzo, list) else fzo.value


def _get_value(expr):
    """ Get the python value of an expression
    Args:
        expr: Expression (python or CPO)
    Returns:
        Python value
    """
    return expr.value if isinstance(expr, CpoValue) else expr


def _build_domain(v):
    """ Build a variable domain from an initial value
    Args:
        v: Variable initial value
    Returns:
        Variable domain
    """
    if v is True:
        return (1,)
    if v is False:
        return (0,)
    return (v,)


# def _build_cpo_value(v):
#     """ Build a value from a single Python value
#     Args:
#         v: Value
#     Returns:
#         Corresponding CPO value
#     """
#     if v is True:
#         v = 1
#     elif v is False:
#         v = 0
#     return build_cpo_expr(v)


def _mutl_by_int(expr, val):
    """ Create an expression that multiply an expression by an integer
    Args:
        expr: Expression to constrain
        val:  Integer value to multiply expression with
    Returns:
        New expression
    """
    if val == 1:
        return expr
    if val == -1:
        return -expr
    if val == 0:
        return 0
    return val * expr





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2020
# --------------------------------------------------------------------------
'''Some csv utilities
'''
import os


def encode_csv_string(text):
    """ Encode a string to be used in CSV file

    Args:
        text:  String to encode
    Returns:
        Encoded string, including starting and ending double quote
    """
    res = ['"']
    for c in text:
        res.append(c)
        if c == '"':
            res.append('"')
    res.append('"')
    return ''.join(res)


def write_csv_line(output, line, encoding):
    line = ','.join([encode_csv_string('%s' % c) for c in line])
    output.write(line.encode(encoding))
    output.write('\n'.encode(encoding))


def write_csv(env, table, fields, name):
    # table must be a named tuple
    encoding = 'utf-8'
    with env.get_output_stream(name) as ostr:
        write_csv_line(ostr, fields, encoding)
        for line in table:
            write_csv_line(ostr, line, encoding)


def write_table_as_csv(env, table, name, field_names):
    '''Writes a kpis dataframe as file which name is specified.
    The data type depends of extension of name.

    This uses the specfied env to write data as attachments
    '''
    _, ext = os.path.splitext(name)
    ext = ext.lower()
    if ext == '.csv':
        encoding = 'utf-8'
        with env.get_output_stream(name) as ostr:
            write_csv_line(ostr, field_names, encoding)
            for line in table:
                write_csv_line(ostr, line, encoding)
    else:
        # right now, only csv is supported
        raise ValueError('file format not supported for KPIs file: %s' % ext)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------


# gendoc: ignore

from docplex.mp.basic import Expr
from docplex.mp.constants import SOSType
from docplex.mp.operand import LinearOperand

from docplex.mp.utils import is_iterable, is_iterator, DocplexLinearRelaxationError

# do NOT import Model -> circular

# change this flag to generate named objects
# by default all generated objects will have no name
use_debug_names = False


def get_name_if_debug(name):
    return name if use_debug_names else None


# noinspection PyAbstractClass
class _FunctionalExpr(Expr, LinearOperand):
    # INTERNAL class
    # parent class for all nonlinear expressions.
    __slots__ = ('_f_var', '_resolved')

    def __init__(self, model, name=None):
        Expr.__init__(self, model, name)
        self._f_var = None
        self._resolved = False

    def to_linear_expr(self):
        return self._get_resolved_f_var()

    def iter_terms(self):
        yield self._get_resolved_f_var(), 1

    iter_sorted_terms = iter_terms

    def iter_variables(self):
        # do we need to create it here?
        yield self._get_resolved_f_var()

    def unchecked_get_coef(self, dvar):
        return 1 if dvar is self._f_var else 0

    def _new_generated_free_continuous_var(self, name=None):
        # INTERNAL
        inf = self._model.infinity
        return self._new_generated_continuous_var(lb=-inf, ub=inf, name=name)

    def _new_generated_continuous_var(self, lb=None, ub=None, name=None):
        return self._new_generated_var(vartype=self._model.continuous_vartype, lb=lb, ub=ub, name=name)

    def _new_generated_binary_var(self, name=None):
        return self._new_generated_var(self._model.binary_vartype, name=name)

    def _new_generated_var(self, vartype, lb=None, ub=None, name=None):
        # INTERNAL
        m = self._model
        gvar = m._lfactory.new_safe_var(vartype, lb=lb, ub=ub, varname=name)
        gvar.notify_origin(self)
        return gvar

    def _new_generated_binary_varlist(self, keys, name=None):
        bvars = self.model.binary_var_list(keys, name)
        for bv in bvars:
            bv.notify_origin(self)
        return bvars

    def new_generated_sos1(self, dvars):
        sos1 = self.model._add_sos(dvars, SOSType.SOS1)
        sos1.notify_origin(self)
        return sos1

    def _new_generated_indicator(self, binary_var, linear_ct, active_value=1, name=None):
        ind = self._model._lfactory.new_indicator_constraint(binary_var, linear_ct, active_value, name)
        ind.notify_origin(self)
        self._model.add(ind)
        return ind

    def _new_generated_binary_ct(self, lhs, rhs, sense='EQ'):
        # posts a constraint and marks it as generated.
        m = self._model
        ct = m._lfactory.new_binary_constraint(lhs=lhs, sense=sense, rhs=rhs)
        m._post_constraint(ct)
        ct.notify_origin(self)
        return ct

    def _post_generated_cts(self, cts):
        # takes a sequence of constraints
        # posts a constraint and marks it as generated.
        self._model._lfactory._post_constraint_block(cts)
        for c in cts:
            c.notify_origin(self)
        return cts

    def _get_resolved_f_var(self):
        self._ensure_resolved()
        return self._f_var

    def _get_allocated_f_var(self):
        if self._f_var is None:
            self._f_var = self._create_functional_var()
        return self._f_var

    def resolve(self):
        self._ensure_resolved()

    def _ensure_resolved(self):
        if self._f_var is None:
            # 1. create the var (once!)
            self._f_var = self._create_functional_var()
            # 2. post the link between the fvar and the argument expr
        if not self._resolved:
            self._resolve()
            self._resolved = True

    def _is_resolved(self):
        return self._resolved and self._f_var is not None

    def _name_functional_var_name(self, fvar, fvar_meta_format="_%s%d"):
        fname = fvar_meta_format % (self.function_symbol, fvar._index)
        fvar.set_name(fname)

    def _create_functional_var(self, named=True):
        fvar = self._new_generated_free_continuous_var(name=None)
        if named:
            self._name_functional_var_name(fvar)
        return fvar

    @property
    def functional_var(self):
        return self._get_resolved_f_var()

    as_var = functional_var

    def square(self):
        return self.functional_var.square()

    def _resolve(self):
        raise NotImplementedError  # pragma: no cover

    def _get_function_symbol(self):
        # redefine this to get the function symbol
        raise NotImplementedError  # pragma: no cover

    @property
    def function_symbol(self):
        return self._get_function_symbol()

    def relaxed_copy(self, relaxed_model, var_map):
        raise DocplexLinearRelaxationError(self, cause=self.function_symbol)

    def __str__(self):
        return self.to_string()

    def to_string(self, **kwargs):
        raise NotImplementedError  # pragma: no cover

    # -- arithmetic operators
    def __mul__(self, e):
        return self.functional_var.__mul__(e)

    def __rmul__(self, e):
        return self.functional_var.__mul__(e)

    def __div__(self, e):
        return self.divide(e)

    def __truediv__(self, e):
        # for py3
        # INTERNAL
        return self.divide(e)  # pragma: no cover

    def divide(self, e):
        return self.functional_var.divide(e)

    def __add__(self, e):
        return self.functional_var.__add__(e)

    def __radd__(self, e):
        return self.functional_var.__add__(e)

    def __sub__(self, e):
        return self.functional_var.__sub__(e)

    def __rsub__(self, e):
        return self.functional_var.__rsub__(e)

    def __neg__(self):
        # the "-e" unary minus returns a linear expression
        return self.functional_var.__neg__()

    def _allocate_arg_var_if_necessary(self, arg_expr):
        # INTERNAL
        # allocates a new variables if only the argument expr is not a variable
        # and returns it
        try:

            arg_var = arg_expr.as_variable()
        except AttributeError:
            arg_var = None

        if arg_var is None:
            arg_var = self._new_generated_free_continuous_var()
            self._new_generated_binary_ct(arg_var, arg_expr)
        return arg_var


# noinspection PyAbstractClass
class UnaryFunctionalExpr(_FunctionalExpr):
    def __init__(self, model, argument_expr, name=None):
        _FunctionalExpr.__init__(self, model, name)
        self._argument_expr = model._lfactory._to_linear_operand(argument_expr)
        self._x_var = self._allocate_arg_var_if_necessary(argument_expr)

    @property
    def argument_expr(self):
        return self._argument_expr

    def is_discrete(self):
        return self._argument_expr.is_discrete()

    def to_string(self):
        return "{0:s}({1!s})".format(self.function_symbol, self._argument_expr)


class AbsExpr(UnaryFunctionalExpr):
    def copy(self, target_model, var_mapping):
        copied_arg_expr = self._argument_expr.copy(target_model, var_mapping)
        return AbsExpr(model=target_model, argument_expr=copied_arg_expr)

    def relaxed_copy(self, relaxed_model, var_map):
        raise DocplexLinearRelaxationError(self, cause='abs')

    def __init__(self, model, argument_expr):
        UnaryFunctionalExpr.__init__(self, model, argument_expr)

    def _get_function_symbol(self):
        return "abs"

    def clone(self):
        return AbsExpr(self.model, self._argument_expr)

    # noinspection PyArgumentEqualDefault,PyArgumentEqualDefault
    def _resolve(self):
        self_f_var = self._f_var
        assert self_f_var
        abs_index = self_f_var.index
        abs_names = ["_abs_pp_%d" % abs_index, "_abs_np_%d" % abs_index] if use_debug_names else [None, None]
        # 1. allocate two variables in one pass.
        positive_var = self._new_generated_continuous_var(lb=0, name=abs_names[0])
        negative_var = self._new_generated_continuous_var(lb=0, name=abs_names[1])

        # F(x) = p + n
        ct1 = (self_f_var == positive_var + negative_var)
        # sos
        self.sos = self.new_generated_sos1(dvars=[positive_var, negative_var])
        # # x = p-n
        ct2 = (self._argument_expr == positive_var - negative_var)

        self._post_generated_cts([ct1, ct2])
        # store
        self.positive_var = positive_var
        self.negative_var = negative_var

    def _get_solution_value(self, s=None):
        raw = abs(self._argument_expr._get_solution_value(s))
        return self._round_if_discrete(raw)

    def __repr__(self):
        return "docplex.mp.AbsExpr({0:s})".format(self._argument_expr.truncated_str())


# noinspection PyAbstractClass
class _SequenceExpr(_FunctionalExpr):
    # INTERNAL: base class for functional exprs with a sequence argument (e.g. min/max)

    def __init__(self, model, exprs, name=None):
        _FunctionalExpr.__init__(self, model, name)
        if is_iterable(exprs) or is_iterator(exprs):
            self._exprs = exprs
        else:
            self._exprs = [model._lfactory._to_linear_operand(exprs)]
        # allocate xvars iff necessary
        self._xvars = [self._allocate_arg_var_if_necessary(e) for e in self._exprs]

    def is_discrete(self):
        return all(map(lambda ex: ex.is_discrete(), self._exprs))

    def _get_args_string(self, sep=","):
        return sep.join(e.truncated_str() for e in self._exprs)

    def to_string(self):
        # generic: format expression arguments with holophraste
        str_args = self._get_args_string()
        return "{0}({1!s})".format(self.function_symbol, str_args)

    def iter_exprs(self):
        return iter(self._exprs)

    def _generate_variables(self):
        # INTERNAL: variable generator scanning all expressions
        # may return the same variable twice (or more)
        # use varset() if you need the set.
        for e in self._exprs:
            for v in e.iter_variables():
                yield v
        yield self._get_resolved_f_var()

    def iter_variables(self):
        return self._generate_variables()

    def contains_var(self, dvar):
        return dvar is self._f_var

    def _get_solution_value(self, s=None):
        fvar = self._f_var
        if self._is_resolved() and (not s or fvar in s):
            raw = fvar._get_solution_value(s)
        else:
            raw = self.compute_solution_value(s)
        return self._round_if_discrete(raw_value=raw)

    def compute_solution_value(self, s):
        raise NotImplementedError  # pragma: no cover

    def copy(self, target_model, var_mapping):
        copied_exprs = [expr.copy(target_model, var_mapping) for expr in self._exprs]
        return self.__class__(target_model, copied_exprs, self.name)

    def clone(self):
        # generic clone
        return self.__class__(self.model, self._exprs, self.name)


class MinimumExpr(_SequenceExpr):
    """ An expression that represents the minimum of a sequence of expressions.

    This expression can be used in all arithmetic operations.
    After a solve, the value of this expression is equal to the minimum of the values
    of its argument expressions.
    """

    def __init__(self, model, exprs, name=None):
        _SequenceExpr.__init__(self, model, exprs, name)

    def _get_function_symbol(self):
        return "min"

    def __repr__(self):
        str_args = self._get_args_string()
        return "docplex.mp.MinExpr({0!s})".format(str_args)

    def _resolve(self):
        self_min_var = self._f_var
        assert self_min_var
        self_x_vars = self._xvars
        nb_args = len(self_x_vars)
        if 0 == nb_args:
            self._f_var.set_bounds(0, 0)
        elif 1 == nb_args:
            self._new_generated_binary_ct(self_min_var, self._xvars[0])
        else:
            cts = []
            for xv in self_x_vars:
                cts.append(self_min_var <= xv)
            # allocate N _generated_ binaries
            z_vars = self._new_generated_binary_varlist(keys=nb_args)
            # sos?
            cts.append(self.model.sum(z_vars) == 1)
            self._post_generated_cts(cts)
            # indicators
            for i in range(nb_args):
                z = z_vars[i]
                x = self_x_vars[i]
                # need a block generation of indicators
                self._new_generated_indicator(binary_var=z, linear_ct=(self_min_var >= x))

    def compute_solution_value(self, s):
        return min(expr._get_solution_value(s) for expr in self._exprs)


class MaximumExpr(_SequenceExpr):
    """ An expression that represents the maximum of a sequence of expressions.

    This expression can be used in all arithmetic operations.
    After a solve, the value of this expression is equal to the minimum of the values
    of its argument expressions.
    """

    def __init__(self, model, exprs, name=None):
        _SequenceExpr.__init__(self, model, exprs, name)

    def _get_function_symbol(self):
        return "max"

    def __repr__(self):
        str_args = self._get_args_string()
        return "docplex.mp.MaxExpr({0!s})".format(str_args)

    def _resolve(self):
        self_max_var = self._f_var
        self_x_vars = self._xvars
        nb_args = len(self_x_vars)
        if 0 == nb_args:
            self._f_var.set_bounds(0, 0)  # what else ??
        elif 1 == nb_args:
            self._new_generated_binary_ct(self_max_var, self._xvars[0])
        else:
            for xv in self_x_vars:
                self._new_generated_binary_ct(self_max_var, xv, 'GE')
            # allocate N binaries
            z_vars = self._new_generated_binary_varlist(keys=nb_args)
            # sos?
            self._new_generated_binary_ct(self.model.sum(z_vars), 1)
            # indicators
            for i in range(nb_args):
                z = z_vars[i]
                x = self_x_vars[i]
                self._new_generated_indicator(binary_var=z, linear_ct=(self_max_var <= x))

    def compute_solution_value(self, s):
        return max(expr._get_solution_value(s) for expr in self._exprs)


class LogicalNotExpr(UnaryFunctionalExpr):
    def _create_functional_var(self, named=True):
        # the resulting variable is a binary variable...
        bvar = self._new_generated_binary_var(name=None)
        self._name_functional_var_name(bvar)
        return bvar

    def is_discrete(self):
        return True

    def _get_function_symbol(self):
        return "not"

    def as_logical_operand(self):
        return self._get_resolved_f_var()

    def copy(self, target_model, var_mapping):
        copied_arg_expr = self._argument_expr.copy(target_model, var_mapping)
        return LogicalNotExpr(model=target_model, argument_expr=copied_arg_expr)

    def __init__(self, model, argument_expr):
        UnaryFunctionalExpr.__init__(self, model, argument_expr)
        self._logical_op_arg = argument_expr.as_logical_operand()
        assert self._logical_op_arg is not None
        self._actual_arg_s = str(argument_expr)

    def to_string(self):
        return "{0:s}({1!s})".format(self.function_symbol, self._actual_arg_s)

    def clone(self):
        return LogicalNotExpr(self.model, self._argument_expr)

    # noinspection PyArgumentEqualDefault,PyArgumentEqualDefault
    def _resolve(self):
        not_var = self._f_var
        assert not_var

        # not_x + x == 1
        ct1 = (not_var + self._logical_op_arg == 1)
        self._post_generated_cts([ct1])
        # store
        self.not_ct = ct1

    def _get_solution_value(self, s=None):
        arg_val = self._argument_expr._get_solution_value(s)
        return 0 if arg_val else 1

    def __repr__(self):
        return "docplex.mp.NotExpr({0:s})".format(self._argument_expr.truncated_str())


class _LogicalSequenceExpr(_SequenceExpr):

    def as_logical_operand(self):
        return self._get_resolved_f_var()

    def _create_functional_var(self, named=True):
        # the resulting variable is a binary variable...
        bvar = self._new_generated_binary_var(name=None)
        self._name_functional_var_name(bvar)
        return bvar

    def __init__(self, model, exprs, name=None):
        _FunctionalExpr.__init__(self, model, name)
        assert is_iterable(exprs) or is_iterator(exprs)
        self._exprs = exprs
        # never allocate vars: arguments --are-- binary variables.
        self._xvars = exprs

    def _get_args_string(self, sep=","):
        s = sep.join(str(b.origin()) if b.is_generated() else str(b) for b in self._xvars)
        return s



    def is_discrete(self):
        return True

    precision = 1e-5


class LogicalAndExpr(_LogicalSequenceExpr):

    def _get_function_symbol(self):
        return "and"

    def __repr__(self):
        str_args = self._get_args_string()
        return "docplex.mp.LogicalAndExpr({0!s})".format(str_args)

    def compute_solution_value(self, s):
        # return 1/0 not True/False
        threshold = 1 - self.precision
        return 1 if all(ex._get_solution_value(s) >= threshold for ex in self._exprs) else 0

    def _resolve(self):
        self_and_var = self._f_var
        self_x_vars = self._xvars

        if self_x_vars:
            cts = [(self_and_var <= xv) for xv in self_x_vars]
            m = self._model
            nb_vars = len(self_x_vars)
            # rtc-39600: subtract n-1 from the sum.
            # the -and- var is propagated to 1 if all sum vars are 1.
            cts.append(self_and_var >= m._aggregator._sum_with_seq(self._xvars) - (nb_vars-1))
            self._post_generated_cts(cts)


class LogicalOrExpr(_LogicalSequenceExpr):

    def _get_function_symbol(self):
        return "or"

    def __repr__(self):
        str_args = self._get_args_string()
        return "docplex.mp.LogicalOrExpr({0!s})".format(str_args)

    def compute_solution_value(self, s):
        # return 1/0 not True/False
        threshold = 1 - self.precision
        return 1 if any(ex._get_solution_value(s) >= threshold for ex in self._exprs) else 0

    def _resolve(self):
        self_or_var = self._f_var
        self_x_vars = self._xvars

        if self_x_vars:
            cts = [(xv <= self_or_var) for xv in self_x_vars]
            m = self._model
            cts.append(self_or_var <= m._aggregator._sum_with_seq(self._xvars))
            self._post_generated_cts(cts)
        self._resolved = True


class PwlExpr(UnaryFunctionalExpr):

    def __init__(self, model, pwl_func, argument_expr, usage_counter, y_var=None,  add_counter_suffix=True, resolve=True):
        UnaryFunctionalExpr.__init__(self, model, argument_expr)
        self._pwl_func = pwl_func
        self._usage_counter = usage_counter
        self._f_var = y_var
        if pwl_func.name:
            # ?
            if add_counter_suffix:
                self.name = '{0}_{1!s}'.format(self._pwl_func.name, self._usage_counter)
            else:
                self.name = self._pwl_func.name
        if resolve:
            self._ensure_resolved()

    def _get_function_symbol(self):
        # this method determines the name of the generated variable
        # as usual it starts with "_" to mark this is a generated variable.
        pwl_name = self._pwl_func.get_name()
        # TODO: what if pwl_name is not LP-compliant??
        return "pwl" if not pwl_name else "pwl_%s#" % pwl_name

    def _get_solution_value(self, s=None):
        raw = self._f_var._get_solution_value(s)
        return self._round_if_discrete(raw)

    def iter_variables(self):
        for v in self._argument_expr.iter_variables():
            yield v
        yield self._get_resolved_f_var()

    def _resolve(self):
        mdl = self._model
        pwl_constraint = mdl._lfactory.new_pwl_constraint(self, self.get_name())
        mdl._add_pwl_constraint_internal(pwl_constraint)

    @property
    def pwl_func(self):
        return self._pwl_func

    @property
    def usage_counter(self):
        return self._usage_counter

    def __repr__(self):
        return "docplex.mp.PwlExpr({0:s}, {1:s})".format(self._get_function_symbol(),
                                                         self._argument_expr.truncated_str())

    def copy(self, target_model, var_map):
        copied_pwl_func = var_map[self.pwl_func]
        copied_x_var = var_map[self._x_var]
        copied_pwl_expr = PwlExpr(target_model, copied_pwl_func, copied_x_var, self.usage_counter)
        copied_pwl_expr_f_var = var_map.get(self._f_var)
        if copied_pwl_expr_f_var:
            copied_pwl_expr._f_var = copied_pwl_expr_f_var
            # Need to set the _origin attribute of the copied var
            copied_pwl_expr_f_var._origin = copied_pwl_expr
        return copied_pwl_expr

    def relaxed_copy(self, relaxed_model, var_map):
        raise DocplexLinearRelaxationError(self, cause='pwl')



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017
# --------------------------------------------------------------------------
# gendoc: ignore

from docplex.mp.advmodel import AdvModel
from docplex.mp.sktrans.pd_utils import make_solution


class DOcplexModeler(object):

    def __init__(self):
        pass

    @classmethod
    def _solve_model(cls, mdl, cols, colnames, objsense, costs, solution_maker=make_solution, **params):
        if costs is not None:
            mdl.set_objective(sense=objsense, expr=mdl.scal_prod_vars_all_different(cols, costs))


        # --- lp export
        lp_export = params.pop('lp_export', False)
        lp_base = params.pop('lp_basename', None)
        lp_path = params.pop('lp_path', None)
        if lp_export:
            mdl.export_as_lp(basename=lp_base, path=lp_path)
        # ---
        keep_zeros = params.pop('keep_zeros', False)
        #mdl.prettyprint()


        # -- solving
        s = mdl.solve()
        all_values = s.get_all_values() if s else None
        return solution_maker(all_values, colnames, keep_zeros)


    @classmethod
    def build_matrix_linear_model_and_solve(cls, var_count, var_lbs, var_ubs,
                                            var_types, var_names,
                                            cts_mat, rhs,
                                            objsense, costs,
                                            cast_to_float,
                                            solution_maker=make_solution,
                                            **transform_params):
        with AdvModel(name='lp_transformer') as mdl:
            varlist = mdl.continuous_var_list(var_count, lb=var_lbs, ub=var_ubs, name=var_names)
            sense = transform_params.get('sense', 'le')
            mdl.add(mdl.matrix_constraints(cts_mat, varlist, rhs, sense=sense))
            return cls._solve_model(mdl, varlist, var_names, objsense, costs=costs, solution_maker=solution_maker,
                                    **transform_params)

    @classmethod
    def build_matrix_range_model_and_solve(cls, var_count, var_lbs, var_ubs,
                                           var_types, var_names,
                                           cts_mat, range_mins, range_maxs,
                                           objsense, costs,
                                           cast_to_float,
                                           solution_maker=make_solution,
                                           **transform_params):
        with AdvModel(name='lpr_transformer') as mdl:
            varlist = mdl.continuous_var_list(var_count, lb=var_lbs, ub=var_ubs, name=var_names)
            mdl.add(mdl.matrix_ranges(cts_mat, varlist, range_mins, range_maxs))
            return cls._solve_model(mdl, varlist, var_names, objsense, costs=costs, solution_maker=solution_maker,
                                    **transform_params)

    @classmethod
    def build_sparse_linear_model_and_solve(cls, nb_vars, var_lbs, var_ubs,
                                            var_types, var_names,
                                            nb_rows, cts_sparse_coefs,
                                            objsense, costs,
                                            solution_maker=make_solution,
                                            **transform_params):
        with AdvModel(name='lp_transformer') as mdl:
            varlist = mdl.continuous_var_list(nb_vars, lb=var_lbs, ub=var_ubs)
            lfactory = mdl._lfactory
            r_rows = range(nb_rows)
            exprs = [lfactory.linear_expr() for _ in r_rows]
            rhss = [0] * nb_rows
            for coef, row, col in cts_sparse_coefs:
                if col < nb_vars:
                    exprs[row]._add_term(varlist[col], coef)
                else:
                    assert col == nb_vars
                    rhss[row] = coef
            sense = transform_params.get('sense', 'le')
            cts = [lfactory.new_binary_constraint(exprs[r], rhs=rhss[r], sense=sense) for r in r_rows]
            lfactory._post_constraint_block(cts)
            return cls._solve_model(mdl, varlist, var_names, objsense=objsense, costs=costs,
                                    solution_maker=solution_maker, **transform_params)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------
from docplex.mp.basic import ModelingObject, ModelingObjectBase, Priority, _BendersAnnotatedMixin
from docplex.mp.constants import ComparisonType, UpdateEvent
from docplex.mp.operand import LinearOperand
from docplex.mp.sttck import StaticTypeChecker
from docplex.mp.utils import DocplexLinearRelaxationError

import warnings


class _ExtraConstraintUsage(object):
    # INTERNAL
    def __init__(self, tag):
        self.tag = tag

    def __str__(self):
        return "UsedAs<%s>" % self.tag

    def notify_expr_modified(self, linct, new_expr, engine):
        tagval = self.tag
        engine.update_extra_constraint(linct, tagval, new_expr)


class _ConstraintLogicalUsage(object):
    def __init__(self, logct):
        self._log_ct = logct

    @property
    def tag(self):
        return "logical"

    _cannot_modify_linearct_non_discrete_msg = 'Linear constraint: {0} is used in equivalence, cannot be modified with non-discrete expr: {1}'

    def notify_expr_modified(self, linct, new_expr, engine):
        log_ct = self._log_ct
        if log_ct.is_equivalence() and not new_expr.is_discrete():
            linct.fatal(self._cannot_modify_linearct_non_discrete_msg, linct, new_expr)
        else:
            engine.update_constraint(log_ct, event=UpdateEvent.IndicatorLinearConstraint)


class AbstractConstraint(ModelingObject, _BendersAnnotatedMixin):
    __slots__ = ()

    def __init__(self, model, name=None):
        ModelingObject.__init__(self, model, name)

    @property
    def priority(self):
        return self._model.get_constraint_priority(self)

    @priority.setter
    def priority(self, newprio):
        self.set_priority(newprio)

    def set_priority(self, newprio):
        self._model.set_constraint_priority(self, Priority.parse(newprio, logger=self.error_handler, accept_none=True))

    def set_mandatory(self):
        ''' Sets the constraint as mandatory.

        This prevents relaxation from relaxing this constraint.
        To revert this, set the priority to any non-mandatory priority, or None.
        '''
        self.priority = Priority.MANDATORY

    def is_mandatory(self):
        return Priority.MANDATORY == self.priority

    def _unsupported_relational_op(self, op_string, other):
        self.fatal("Relational operator: {1} is unavailable with constraint: {0!s}", self, op_string)

    def __le__(self, e):
        self._unsupported_relational_op("<=", e)

    def __ge__(self, e):
        self._unsupported_relational_op(">=", e)

    def __lt__(self, e):
        self._unsupported_relational_op("<", e)

    def __gt__(self, e):
        self._unsupported_relational_op(">", e)

    # py3 needs an eq for hashing
    # def __eq__(self, e):
    #     self._unsupported_relational_op("==", e)

    def _no_linear_ct_in_logical_test_error(self):
        raise TypeError("cannot convert a constraint to boolean: {0!s}".format(self))

    def __nonzero__(self):
        self._no_linear_ct_in_logical_test_error()

    def __bool__(self):
        # python 3 version of nonzero
        self._no_linear_ct_in_logical_test_error()  # pragma: no cover

    def iter_variables(self):
        raise NotImplementedError  # pragma: no cover

    def iter_exprs(self):
        raise NotImplementedError  # pragma: no cover

    def get_var_coef(self, dvar):  # pragma: no cover
        raise NotImplementedError

    @property
    def size(self):
        return sum(x.size for x in self.iter_exprs())

    def copy(self, target_model, var_map):
        raise NotImplementedError  # pragma: no cover

    def relaxed_copy(self, relaxed_model, var_map):
        return self.copy(relaxed_model, var_map)

    def compute_infeasibility(self, slack):  # pragma: no cover
        # INTERNAL: only used when json has no infeasibility info.
        return slack

    # noinspection PyMethodMayBeStatic
    def notify_deleted(self):
        # INTERNAL
        self._set_invalid_index()

    @property
    def short_typename(self):
        return "constraint"

    def is_trivial(self):
        return False

    def is_linear(self):
        return False

    def is_quadratic(self):
        return False

    def is_added(self):
        """ Returns True if the constraint has been added to its model.

        Example:
            c = (x+y == 1)
            m.add(c)
            c.is_added()
            >>> True
            c2 = (x + 2*y) >= 3
            c2.is_added()
            >>> False

        """
        return self.index >= 0

    @property
    def cplex_scope(self):
        return self._get_index_scope().cplex_scope

    def _get_index_scope(self):
        raise NotImplementedError

    def is_logical(self):
        return False

    def _get_dual_value(self):
        # INTERNAL
        # Note that dual values are only available for LP problems,
        # so can be calle donly on linear or range constraints.
        return self._model._dual_value1(self)

    def notify_expr_modified(self, expr, event):
        # INTERNAL
        pass  # pragma: no cover

    def notify_expr_replaced(self, expr, new_expr):
        # INTERNAL
        pass  # pragma: no cover

    def resolve(self):
        raise NotImplementedError  # pragma: no cover

    def is_satisfied(self, solution, tolerance):
        raise NotImplementedError  # pragma: no cover

    def lock_discrete(self):
        # lock sub expressions
        for expr in self.iter_exprs():
            expr.lock_discrete()


# noinspection PyAbstractClass
class BinaryConstraint(AbstractConstraint):
    __slots__ = ("_ctsense", "_left_expr", "_right_expr")

    def _internal_set_sense(self, new_sense):
        self._ctsense = new_sense

    def __init__(self, model, left_expr, ctsense, right_expr, name=None):
        ModelingObject.__init__(self, model, name)
        self._ctsense = ctsense
        # noinspection PyPep8
        self._left_expr = left_expr
        self._right_expr = right_expr

    def _iter_usages(self):
        return iter([])

    @property
    def type(self):
        """ This property returns the type of the constraint; type is an enumerated value
        of type :class:`ComparisonType`, with three possible values:

        - LE for e1 <= e2 constraints

        - EQ for e1 == e2 constraints

        - GE for e1 >= e2 constraints

        where e1 and e2 denote linear expressions.

        """
        return self._ctsense

    def cplex_code(self):
        return self._ctsense._cplex_code

    def get_left_expr(self):
        """ This property returns the left expression in the constraint.

        Example:
            (X+Y <= Z+1) has left expression (X+Y).
        """
        return self._left_expr

    def get_right_expr(self):
        """ This property returns the right expression in the constraint.

        Example:
            (X+Y <= Z+1) has right expression (Z+1).
        """
        return self._right_expr

    def get_var_coef(self, dvar):
        return self._left_expr.unchecked_get_coef(dvar) - self._right_expr.unchecked_get_coef(dvar)

    def to_string(self):
        """ Returns a string representation of the constraint.

        The operators in this representation are the usual operators <=, ==, and >=.

        Example:
            The constraint (X+Y <= Z+1) is represented as "X+Y <= Z+1".

        Returns:
            A string.

        """
        left_string = self._left_expr.to_string()
        right_string = self._right_expr.to_string()
        self_name = self.name
        if self_name:
            return u"%s: %s %s %s" % (self_name,
                                      left_string,
                                      self._ctsense.operator_symbol,
                                      right_string)
        else:
            return u"%s %s %s" % (left_string,
                                  self._ctsense.operator_symbol,
                                  right_string)

    def cplex_num_rhs(self):
        # INTERNAL
        right_cst = self._right_expr.get_constant()
        left_cst = self._left_expr.get_constant()
        return float(right_cst - left_cst)

    def __repr__(self):
        classname = self.__class__.__name__
        user_name = self.safe_name
        typename = self._ctsense.short_name
        sleft = self._left_expr.truncated_str()
        sright = self._right_expr.truncated_str()
        return "docplex.mp.{0}[{1}]({2!s},{3},{4!s})". \
            format(classname, user_name, sleft, typename, sright)

    def _is_trivially_feasible(self):
        # INTERNAL : assume self is trivial()
        op_func = self._ctsense.python_operator
        return op_func(self._left_expr.get_constant(), self._right_expr.get_constant()) if op_func else False

    def _is_trivially_infeasible(self):
        # INTERNAL: assume self is trivial .
        op_func = self._ctsense.python_operator
        return not op_func(self._left_expr.get_constant(), self._right_expr.get_constant()) if op_func else False

    def is_trivial_feasible(self):
        return self.is_trivial() and self._is_trivially_feasible()

    def is_trivial_infeasible(self):
        return self.is_trivial() and self._is_trivially_infeasible()

    @staticmethod
    def _generate_expr_opposite_linear_coefs(expr):
        for v, k in expr.iter_sorted_terms():
            yield v, -k

    def _iter_net_linear_coefs_sorted(self, left_expr, right_expr):
        # INTERNAL
        if right_expr.is_constant():
            return left_expr.iter_sorted_terms()
        elif left_expr.is_constant():
            return self._generate_expr_opposite_linear_coefs(right_expr)
        else:
            return self._generate_net_linear_coefs2_sorted(left_expr, right_expr)

    def iter_variables(self):
        """  Iterates over all variables mentioned in the constraint.

        *Note:* This includes variables that are mentioned with a zero coefficient. For example,
        the iterator on the following constraint:

         X <= X+Y + 1

        will return X and Y, although X is mentioned with a zero coefficient.

        Returns:
            An iterator object.
        """
        if self._right_expr.is_constant():
            return self._left_expr.iter_variables()
        elif self._left_expr.is_constant():
            return self._right_expr.iter_variables()
        else:
            return self.generate_ordered_vars()

    def generate_ordered_vars(self):
        left_expr = self._left_expr
        for lv in left_expr.iter_variables():
            yield lv
        for rv in self._right_expr.iter_variables():
            if not left_expr.contains_var(rv):
                yield rv

    @staticmethod
    def _generate_net_linear_coefs2_sorted(left_expr, right_expr):
        # INTERNAL
        for lv, lk in left_expr.iter_sorted_terms():
            net_k = lk - right_expr.unchecked_get_coef(lv)
            if net_k:
                yield lv, net_k
        for rv, rk in right_expr.iter_sorted_terms():
            if not left_expr.contains_var(rv) and rk:
                yield rv, -rk

    @staticmethod
    def _generate_net_linear_coefs2_unsorted(left_expr, right_expr):
        # INTERNAL
        for lv, lk in left_expr.iter_terms():
            net_k = lk - right_expr.unchecked_get_coef(lv)
            yield lv, net_k
        for rv, rk in right_expr.iter_terms():
            if not left_expr.contains_var(rv):
                yield rv, -rk

    def notify_deleted(self):
        # INTERNAL
        super(BinaryConstraint, self).notify_deleted()
        self._left_expr.notify_unsubscribed(self)
        self._right_expr.notify_unsubscribed(self)

    def iter_exprs(self):
        return iter([self._left_expr, self._right_expr])

    def get_expr_from_pos(self, pos):
        if 0 == pos:
            return self._left_expr
        elif 1 == pos:
            return self._right_expr
        else:  # pragma: no cover
            self.fatal('Unexpected expression position: {0!r}, expecting 0 or 1', pos)

    def set_expr_from_pos(self, pos, new_expr):
        if 0 == pos:
            self._left_expr = new_expr
        elif 1 == pos:
            self._right_expr = new_expr
        else:  # pragma: no cover
            self.fatal('Unexpected expression position: {0!r}, expecting 0 or 1', pos)

    def is_satisfied(self, solution, tolerance=1e-6):
        left_value = self._left_expr._get_solution_value(solution)
        right_value = self._right_expr._get_solution_value(solution)
        return ComparisonType.almost_compare(left_value, self._ctsense, right_value, eps=tolerance)

    def resolve(self):
        self._left_expr.resolve()
        self._right_expr.resolve()

    def is_discrete(self):
        return self.get_left_expr().is_discrete() and self.get_right_expr().is_discrete()


class LinearConstraint(BinaryConstraint, LinearOperand):
    """ The class that models all constraints of the form `<expr1> <OP> <expr2>`,
            where <expr1> and <expr2> are linear expressions.
    """
    __slots__ = ('_status_var', '_usages')

    def __init__(self, model, left_expr, ctsense, right_expr, name=None):
        BinaryConstraint.__init__(self, model, left_expr, ctsense, right_expr, name)
        left_expr.notify_used(self)
        right_expr.notify_used(self)

    def check_name(self, new_name):
        if new_name is not None:
            self.check_lp_name(new_name)

    def set_name(self, new_name):
        # INTERNAL
        self.check_name(new_name)
        if self.is_added():
            self._model.set_linear_constraint_name(self, new_name)
        else:
            self._set_name(new_name)

    name = property(ModelingObjectBase.get_name, set_name)

    def notify_used_in_logical_ct(self, lct):
        self._add_usage(_ConstraintLogicalUsage(lct))

    def _add_usage(self, usage):
        if hasattr(self, '_usages'):
            self._usages.append(usage)
        else:
            self._usages = [usage]

    def _get_usages(self):
        return getattr(self, '_usages', [])

    def _iter_usages(self):
        # INTERNAL
        return iter(self._get_usages())

    def is_linear(self):
        return True

    @property
    def short_typename(self):
        return "linear constraint"

    def cplex_range_value(self):
        return 0.0

    def is_lazy_constraint(self):
        return self.model._is_lazy_constraint(self)

    def is_user_cut_constraint(self):
        return self.model._is_user_cut_constraint(self)

    def copy(self, target_model, var_map):
        copied_left = self.left_expr.copy(target_model, var_map)
        copied_right = self.right_expr.copy(target_model, var_map)
        copy_name = None if target_model.ignore_names else self.name
        return self.__class__(target_model, copied_left, self.sense, copied_right, copy_name)

    def relaxed_copy(self, target_model, var_map):
        copied_left = self.left_expr.relaxed_copy(target_model, var_map)
        copied_right = self.right_expr.relaxed_copy(target_model, var_map)
        copy_name = self.name
        return self.__class__(target_model, copied_left, self.sense, copied_right, copy_name)

    @property
    def sense(self):
        """ This property is used to get or set the sense of the constraint; sense is an enumerated value
        of type :class:`ComparisonType`, with three possible values:

        - LE for e1 <= e2 constraints

        - EQ for e1 == e2 constraints

        - GE for e1 >= e2 constraints

        where e1 and e2 denote linear expressions.

        """
        return self._ctsense

    @sense.setter
    def sense(self, new_sense):
        self.set_sense(new_sense)

    def set_sense(self, new_sense):
        self.get_linear_factory().set_linear_constraint_sense(self, new_sense)

    # compatibility
    @property
    def type(self):  # pragma: no cover
        warnings.warn(
            "ct.type is deprecated, use ct.sense instead.",
            DeprecationWarning, stacklevel=2)
        return self._ctsense

    @property
    def left_expr(self):
        """ This property returns the left expression in the constraint.

        Example:
            (X+Y <= Z+1) has left expression (X+Y).
        """
        return self._left_expr

    @property
    def right_expr(self):
        """ This property returns the right expression in the constraint.

        Example:
            (X+Y <= Z+1) has right expression (Z+1).
        """
        return self._right_expr

    @right_expr.setter
    def right_expr(self, new_rexpr):
        self.set_right_expr(new_rexpr)

    def set_right_expr(self, new_rexpr):
        self.get_linear_factory().set_linear_constraint_right_expr(ct=self, new_rexpr=new_rexpr)

    @left_expr.setter
    def left_expr(self, new_lexpr):
        self.set_left_expr(new_lexpr)

    def set_left_expr(self, new_lexpr):
        self.get_linear_factory().set_linear_constraint_left_expr(ct=self, new_lexpr=new_lexpr)

    # aliases
    lhs = left_expr
    rhs = right_expr

    def _no_linear_ct_in_logical_test_error(self):
        if self.sense == ComparisonType.EQ:
            # for equality testing there -is- a workaround
            msg = "Cannot use == to test expression equality, try using Python is operator or method equals: {0!s}".format(self)
        else:
            msg = "Cannot convert a linear constraint to boolean: {0!s}".format(self)
        raise TypeError(msg)

    def _cannot_promote_from_linear_to_quadratic(self, old_expr, new_expr):
        msg = 'Cannot change linear constraint expr from linear to quadratic'
        if new_expr is None:
            self.fatal('{0}: {1}', msg, old_expr)
        else:
            self.fatal('{0}: was: {1!s}, new: {2!s}', msg, old_expr, new_expr)

    def _check_editable(self, new_expr, engine):
        for usage in self._iter_usages():
            usage.notify_expr_modified(self, new_expr, engine)

    def notify_expr_modified(self, expr, expr_event):
        # INTERNAL
        if expr_event:
            if expr_event is UpdateEvent.LinExprPromotedToQuad:
                self._cannot_promote_from_linear_to_quadratic(old_expr=expr, new_expr=None)
            else:
                self.get_linear_factory().update_linear_constraint_exprs(ct=self, expr_event=expr_event)

    def notify_expr_replaced(self, old_expr, new_expr):
        # INTERNAL
        if new_expr.is_quad_expr() and not old_expr.is_quad_expr():
            self._cannot_promote_from_linear_to_quadratic(old_expr, new_expr)
        if old_expr is self._left_expr:
            self.get_linear_factory().set_linear_constraint_expr_from_pos(lct=self, pos=0, new_expr=new_expr,
                                                                          update_subscribers=False)
        elif old_expr is self._right_expr:
            self.get_linear_factory().set_linear_constraint_expr_from_pos(lct=self, pos=1, new_expr=new_expr,
                                                                          update_subscribers=False)
        else:
            # should not happen
            pass
        # new expr takes al subscribers from old expr
        new_expr.grab_subscribers(old_expr)

    def to_string(self):
        """ Returns a string representation of the constraint.

        The operators in this representation are the usual operators <=, ==, and >=.

        Example:
            The constraint (X+Y <= Z+1) is represented as "X+Y <= Z+1".

        Returns:
            A string.

        """
        return BinaryConstraint.to_string(self)

    def compute_infeasibility(self, slack):  # pragma: no cover
        ctsense = self._ctsense
        if ctsense == ComparisonType.EQ:
            infeas = slack
        elif ComparisonType.LE == ctsense:
            infeas = slack if slack <= 0 else 0
        elif ComparisonType.GE == ctsense:
            infeas = slack if slack >= 0 else 0
        else:
            infeas = 0
        return infeas

    def _get_index_scope(self):
        return self._model._linct_scope

    def is_trivial(self):
        # Checks whether the constraint is equivalent to a comparison between numbers.
        # For example, x <= x+1 is trivial, but 1.5 X <= X + 1 is not.
        self_left_expr = self._left_expr
        self_right_expr = self._right_expr
        if self_right_expr.is_constant():
            for rv, rk in self_left_expr.iter_terms():
                if rk:
                    return False
            else:
                return True

        elif self_left_expr.is_constant():
            for lv, lk in self_right_expr.iter_terms():
                if lk:
                    return False
            else:
                return True
        else:
            for _, nk in BinaryConstraint._generate_net_linear_coefs2_unsorted(self_left_expr, self_right_expr):
                if nk:
                    return False
            else:
                return True

    def _post_meta_constraint(self, rhs, ctsense):
        status_var = self.get_resolved_status_var()
        return self._model._qfactory.new_xconstraint(lhs=status_var, rhs=rhs, comparaison_type=ctsense)

    def le(self, rhs):
        return self._post_meta_constraint(rhs, ComparisonType.LE)

    def eq(self, rhs):
        return self._post_meta_constraint(rhs, ComparisonType.EQ)

    def ge(self, rhs):
        return self._post_meta_constraint(rhs, ComparisonType.GE)

    def __le__(self, rhs):
        return self.le(rhs)

    def __ge__(self, rhs):
        return self.ge(rhs)

    def __hash__(self):
        return id(self)

    def unchecked_get_coef(self, dvar):
        return 1 if dvar is self._get_status_var() else 0

    def contains_var(self, dvar):
        # only user vars
        return dvar is self._get_status_var()

    def iter_terms(self):
        yield self.get_resolved_status_var(), 1

    iter_sorted_terms = iter_terms

    @property
    def dual_value(self):
        """ This property returns the dual value of the constraint.

        Note:
            This method will raise an exception if the model has not been solved successfully.
            
            This method is OK with small numbers of constraints. For large numbers of constraints (>100),
            consider using Model.dual_values() with a sequence of constraints.
            
        See Also:
            `func:docplex.mp.model.Model.dual_values()`
        """
        return self._get_dual_value()

    @property
    def slack_value(self):
        """ This property returns the slack value of the constraint.

        Note:
            This method will raise an exception if the model has not been solved successfully.
            
                        
            This method is OK with small numbers of constraints. For large numbers of constraints (>100),
            consider using Model.slack_values() with a sequence of constraints.
            
        See Also:
            `func:docplex.mp.model.Model.slack_values()`
        """
        return self._model._slack_value1(self)

    @property
    def basis_status(self):
        """ This property returns the basis status of the slack variable of the constraint, if any.

        Returns:
            An enumerated value from the enumerated type `docplex.constants.BasisStatus`.

        Note:
            for the model to hold basis information, the model must have been solved as a LP problem.
            In some cases, a model which failed to solve may still have a basis available. Use
            `Model.has_basis()` to check whether the model has basis information or not.

        See Also:
            :func:`docplex.mp.model.Model.has_basis`
            :class:`docplex.mp.constants.BasisStatus`

        *New in version 2.10*

        """
        return self._model._linearct_basis_status([self])[0]

    def iter_net_linear_coefs(self):
        # INTERNAL
        left_expr = self._left_expr
        right_expr = self._right_expr
        if right_expr.is_constant():
            return left_expr.iter_sorted_terms()
        elif left_expr.is_constant():
            return self._generate_expr_opposite_linear_coefs(right_expr)
        else:
            return self._generate_net_linear_coefs2_sorted(left_expr, right_expr)

    def _get_status_var(self):
        # this call does -not- create the variable. Returns None if not present.
        return getattr(self, '_status_var', None)

    @property
    def status_var(self):
        return self.get_resolved_status_var()

    as_var = status_var

    def as_logical_operand(self):
        if not self.is_discrete():
            return None
        else:
            return self.get_resolved_status_var\
                (caller_msg='Conversion to logical operand is available only for discrete constraints')

    def _check_is_discrete(self, ct, msg=None):
        err_msg = msg or "Conversion from constraint to expression is available only for discrete constraints"
        StaticTypeChecker.typecheck_discrete_constraint(self, ct, msg=err_msg)

    def get_resolved_status_var(self, caller_msg=None):
        status_var = self._get_status_var()  # always use the getter!
        if status_var is not None:
            return status_var

        self._model.check_logical_constraint_support()

        # TODO: issue a meaningful message on why the ct is not discrete
        self._check_is_discrete(self, msg=caller_msg)
        # lock it discrete
        self.lock_discrete()

        # lazy allocation of a new status variable...:
        lfactory = self.get_linear_factory()
        status_var = lfactory.new_constraint_status_var(self)
        if self.is_added():
            # status variable is bound
            status_var.lb = 1
        self._status_var = status_var
        # store ct in model
        eqct = lfactory.new_equivalence_constraint(status_var, linear_ct=self)
        eqct.notify_origin(self)
        self._model._register_implicit_equivalence_ct(eqct)
        return status_var

    def to_linear_expr(self):
        return self.status_var

    def notify_deleted(self):
        super(LinearConstraint, self).notify_deleted()
        svar = self._get_status_var()  # possibly not resolved
        if svar:
            svar.lb = 0

    @property
    def benders_annotation(self):
        """
        This property is used to get or set the Benders annotation of a constraint.
        The value of the annotation must be a positive integer

        """
        return self.get_benders_annotation()

    @benders_annotation.setter
    def benders_annotation(self, new_anno):
        self.set_benders_annotation(new_anno)

    # -- arithmetic operators
    def times(self, e):
        # TODO: dow e limit numbers here, otherwise non-convex errors may creep in
        return self.to_linear_expr().__mul__(e)

    def __mul__(self, e):
        return self.times(e)

    def __rmul__(self, e):
        return self.times(e)

    def __div__(self, e):
        return self.quotient(e)

    def __truediv__(self, e):
        # for py3
        # INTERNAL
        return self.quotient(e)  # pragma: no cover

    def quotient(self, e):
        svar = self.get_resolved_status_var()
        self._model.typecheck_as_denominator(e, svar)
        inverse = 1.0 / float(e)
        return self.times(inverse)

    def __add__(self, e):
        return self.to_linear_expr().__add__(e)

    def __radd__(self, e):
        return self.to_linear_expr().__add__(e)

    def __sub__(self, e):
        return self.to_linear_expr().__sub__(e)

    def __rsub__(self, e):
        return self.to_linear_expr().__rsub__(e)

    def __or__(self, other):
        return self.logical_or(other)

    def __and__(self, other):
        return self.logical_and(other)

    def _check_logical_operator(self, other, caller=None):
        self._check_is_discrete(self, msg="Logical operators require discrete constraints")
        StaticTypeChecker.typecheck_logical_op(self, other, caller=caller)

    def logical_or(self, other):
        self._check_logical_operator(other, "LinearConstraint.or")
        return self.get_linear_factory().new_binary_constraint_or(self, other)

    def logical_and(self, other):
        self._check_logical_operator(other, "LinearConstraint.and")
        return self.get_linear_factory().new_binary_constraint_and(self, other)

    def __rshift__(self, other):
        """ Redefines the right-shift operator to define if-then constraints.

            This operator allows to create if-then constraint with the `>>` operator.
            It expects a linear constraint as second argument.

            :param other: the linear constraint used to build a new if-then constraint.

            :return:
                an instance of IfThenConstraint, that is not added to the model.
                Use `Model.add()` to add it to the model.

            Note:
                The constraint must be discrete, otherwise an exception is raised.

            Example:

                m.add((x >= 3) >> (y == 5))

                creates  an if-then constraint which links the satisfaction of constraint (x >= 3) to the satisfaction
                of (y ==5).

            """
        return self._model.if_then(if_ct=self, then_ct=other)

    def _tag_as_extra_ct(self, qualifier):
        # INTERNAL
        self._add_usage(_ExtraConstraintUsage(qualifier))

    def _untag_as_extra_ct(self, qualifier):
        usages = self._get_usages()
        upos = -1
        for u, used in enumerate(usages):
            if used.tag == qualifier:
                upos = u
                break
        if upos >= 0:
            del usages[upos]

    _user_cut_tag = "user-cut constraint"
    _lazy_constraint_tag = "lazy constraint"

    def notify_used_as_user_cut(self):
        # INTERNAL
        self._tag_as_extra_ct(self._user_cut_tag)

    def notify_used_as_lazy_constraint(self):
        # INTERNAL
        self._tag_as_extra_ct(self._lazy_constraint_tag)

    def notify_unused_as_user_cut(self):
        # INTERNAL
        self._untag_as_extra_ct(self._user_cut_tag)

    def notify_unused_as_lazy_constraint(self):
        # INTERNAL
        self._untag_as_extra_ct(self._lazy_constraint_tag)



class RangeConstraint(AbstractConstraint):
    """ This class models range constraints.

    A range constraint states that an expression must stay between two
    values, `lb` and `ub`.

    This class is not meant to be instantiated by the user.
    To create a range constraint, use the factory method :func:`docplex.mp.model.Model.add_range`
    defined on :class:`docplex.mp.model.Model`.

    """

    def __init__(self, model, expr, lb, ub, name=None):
        AbstractConstraint.__init__(self, model, name)
        self._ub = ub
        self._lb = lb
        self._expr = expr

    def is_linear(self):
        return True

    # noinspection PyMethodMayBeStatic
    def cplex_code(self):
        return 'R'

    def _get_index_scope(self):
        return self._model._linct_scope

    @property
    def short_typename(self):
        return "range constraint"

    def is_trivial(self):
        return self._expr.is_constant()

    def _is_trivially_feasible(self):
        # INTERNAL : assume self is trivial()
        expr_num = self._expr.constant
        return self._lb <= expr_num <= self._ub

    def _is_trivially_infeasible(self):
        # INTERNAL : assume self is trivial()
        expr_num = self._expr.constant
        return expr_num < self._lb or expr_num > self._ub

    def compute_infeasibility(self, slack):  # pragma: no cover
        # compatible with cplex...
        return -slack

    def get_var_coef(self, dvar):
        return self._expr.unchecked_get_coef(dvar)

    def is_satisfied(self, solution, tolerance=1e-6):
        expr_value = self._expr._get_solution_value(solution)
        return self._lb - tolerance <= expr_value <= self._ub + tolerance

    @property
    def expr(self):
        """ This property returns the linear expression of the range constraint.
        """
        return self._expr

    @expr.setter
    def expr(self, new_expr):
        self.get_linear_factory().set_range_constraint_expr(self, new_expr)

    @property
    def lb(self):
        """ This property is used to get or set the lower bound of the range constraint.

        """
        return self._lb

    @lb.setter
    def lb(self, new_lb):
        self._model.typecheck_num(new_lb)
        self.get_linear_factory().set_range_constraint_lb(self, new_lb)

    @property
    def ub(self):
        """ This property is used to get or set the upper bound of the range constraint.

        """
        return self._ub

    @ub.setter
    def ub(self, new_ub):
        self._model.typecheck_num(new_ub)
        self.get_linear_factory().set_range_constraint_ub(self, new_ub)

    @property
    def bounds(self):
        """ This property is used to get or set the (lower, upper) bounds of a range constraint.

        """
        return self._lb, self._ub

    @bounds.setter
    def bounds(self, new_bounds):
        try:
            new_lb, new_ub = new_bounds
            self._model.typecheck_num(new_lb)
            self._model.typecheck_num(new_ub)
            self.get_linear_factory().set_range_constraint_bounds(self, new_lb, new_ub)
        except ValueError:
            self.fatal('RangeConstraint.bounds expects a 2-tuple of numbers, {0!r} was passed', new_bounds)

    def _internal_set_lb(self, new_lb):
        self._lb = new_lb

    def _internal_set_ub(self, new_ub):
        self._ub = new_ub

    def is_feasible(self):
        return self._ub >= self._lb

    @property
    def dual_value(self):
        """ This property returns the dual value of the constraint.

        Note:
            This method will raise an exception if the model has not been solved successfully.
        """
        return self._get_dual_value()

    @property
    def slack_value(self):
        """ This property returns the slack value of the constraint.

        Note:
            This method will raise an exception if the model has not been solved successfully.
        """
        return self._model._slack_value1(self)

    @property
    def basis_status(self):
        """ This property returns the basis status of the slack variable of the constraint, if any.

        Returns:
            An enumerated value from the enumerated type `docplex.constants.BasisStatus`.

        Note:
            for the model to hold basis information, the model must have been solved as a LP problem.
            In some cases, a model which failed to solve may still have a basis available. Use
            `Model.has_basis()` to check whether the model has basis information or not.

        See Also:
            :func:`docplex.mp.model.Model.has_basis`
            :class:`docplex.mp.constants.BasisStatus`

        *New in version 2.10*
        """
        return self._model._linearct_basis_status([self])[0]

    def iter_variables(self):
        """Iterates over all the variables of the range constraint.

        Returns:
           An iterator object.
        """
        return self._expr.iter_variables()

    def iter_exprs(self):
        yield self._expr

    def cplex_range_value(self, do_raise=True):
        return self.static_cplex_range_value(self, self._lb, self._ub, lambda: "Range has infeasible domain: {0!s}".format(self),
                                             do_raise=do_raise)

    @classmethod
    def static_cplex_range_value(cls, logger, lbval, ubval, msg_fun, do_raise=True):
        rangeval = float(lbval - ubval)
        # this should be negative, otherwise fails....
        # no way to model infeasible ranges with cplex rngval.
        if rangeval >= 1e-6 and do_raise:
            logger.fatal(msg_fun())
        return rangeval

    def cplex_num_rhs(self):
        # force conversion to float for numpy, etc...
        return float(self._ub - self._expr.get_constant())

    def get_left_expr(self):
        return self._expr

    # noinspection PyMethodMayBeStatic
    def get_right_expr(self):
        return None

    def copy(self, target_model, var_map):
        copied_expr = self.expr.copy(target_model, var_map)
        copy_name = None if target_model.ignore_names else self.name
        copied_range = RangeConstraint(target_model, copied_expr, self.lb, self.ub, copy_name)
        return copied_range

    def to_string(self):
        np = self.model._num_printer
        return "{0} <= {1!s} <= {2}".format(np.to_string(self._lb), self._expr, np.to_string(self._ub))

    def __str__(self):
        """ Returns a string representation of the range constraint.

        Example:
            1 <= x+y+z <= 3 represents the range constraint where the expression (x+y+z) is
            constrained to stay between 1 and 3.

        Returns:
            A string.
        """
        return self.to_string()

    def __repr__(self):
        printable_name = self.safe_name
        return "docplex.mp.RangeConstraint[{0}]({1},{2!s},{3})".\
            format(printable_name, self.lb, self._expr, self.ub)

    def resolve(self):
        self._expr.resolve()

    @property
    def benders_annotation(self):
        """
        This property is used to get or set the Benders annotation of a constraint.
        The value of the annotation must be a positive integer

        """
        return self.get_benders_annotation()

    @benders_annotation.setter
    def benders_annotation(self, new_anno):
        self.set_benders_annotation(new_anno)


class NotEqualConstraint(LinearConstraint):

    def __init__(self, model, negated_eqct, name=None):
        # assume negated_eqct is the equality constraint we want to negate
        self._negated_ct = negated_eqct
        aux_eq_ct_status = self._negated_ct.get_resolved_status_var()
        zero = model._lfactory.new_zero_expr()
        LinearConstraint.__init__(self, model, aux_eq_ct_status, ComparisonType.EQ, zero, name)
        self.lock_discrete()

    def to_string(self):
        eqct = self._negated_ct
        return "{0} != {1}".format(eqct._left_expr, eqct._right_expr)

    def set_sense(self, new_sense):
        self.fatal("cannot modify sense of a not_equal constraint: {0!s}", self)

    @property
    def negated_constraint(self):
        return self._negated_ct

    def __str__(self):
        """ Returns a string representation of the not equals constraint.

        Returns:
            A string.
        """
        return self.to_string()

    def __repr__(self):
        return "docplex.mp.NotEquals({0}, {1})". \
            format(self._left_expr, self._right_expr)


class LogicalConstraint(AbstractConstraint):
    """ This class models logical constraints.

    An equivalence constraint links (both ways) the value of a binary variable
    to the satisfaction of a linear constraint.
    
    If the binary variable equals the truth value (default is 1),
    then the constraint is satisfied, conversely if the constraint
    is satisfied, the value of the variable is set to the truth value.

    This class is not meant to be instantiated by the user.

    """
    __slots__ = ('_binary_var', '_linear_ct', '_active_value')

    def __init__(self, model, binary_var, linear_ct, active_value=1, name=None):
        AbstractConstraint.__init__(self, model, name)
        self._binary_var = binary_var
        self._linear_ct = linear_ct
        self._active_value = active_value
        # connect exprs
        for expr in linear_ct.iter_exprs():
            expr.notify_used(self)

        linear_ct.notify_used_in_logical_ct(self)

    def resolve(self):
        self._linear_ct.resolve()

    def _get_index_scope(self):
        return self._model._logical_scope

    def iter_exprs(self):
        return self._linear_ct.iter_exprs()

    def is_equivalence(self):
        raise NotImplementedError  # pragma: no cover

    def cplex_num_rhs(self):
        return self._linear_ct.cplex_num_rhs()

    @property
    def active_value(self):
        return self._active_value

    @property
    def binary_var(self):
        return self._binary_var

    @property
    def linear_constraint(self):
        return self._linear_ct

    @property
    def benders_annotation(self):
        """
        This property is used to get or set the Benders annotation of a constraint.
        The value of the annotation must be a positive integer

        """
        return self.get_benders_annotation()

    @benders_annotation.setter
    def benders_annotation(self, new_anno):
        self.set_benders_annotation(new_anno)

    def get_linear_constraint(self):
        return self._linear_ct

    @property
    def cpx_complemented(self):
        return 1 - self._active_value

    def copy(self, target_model, var_map):
        copied_binary = var_map[self.binary_var]
        copied_linear_ct = self.linear_constraint.copy(target_model, var_map)
        copy_name = None if target_model.ignore_names else self.name
        copied_equiv = self.__class__(target_model,
                                      copied_binary,
                                      copied_linear_ct,
                                      self._active_value,
                                      copy_name)
        return copied_equiv

    def relaxed_copy(self, relaxed_model, var_map):
        raise DocplexLinearRelaxationError(self, cause='logical')

    def is_logical(self):
        return True

    def iter_variables(self):
        yield self._binary_var
        for v in self._linear_ct.iter_variables():
            yield v

    def get_var_coef(self, dvar):
        if dvar is self._binary_var:
            return 1
        else:
            return self._linear_ct.get_var_coef(dvar)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        printable_name = self.safe_name
        clazzname = self.__class__.__name__
        return "docplex.mp.constr.{0:s}[{1}]({2!s},{3!s},true={4})" \
            .format(clazzname, printable_name, self._binary_var, self._linear_ct, self._active_value)

    def notify_expr_modified(self, expr, event):
        # INTERNAL
        self.get_linear_factory().update_indicator_constraint_expr(self, event, expr)

    @property
    def slack_value(self):
        return self._model._slack_value1(self)

    def _symbol(self):
        return '<->' if self.is_equivalence() else '->'

    def to_string(self):
        """
        Displays the equivalence constraint in a (shortened) LP style:
        z <-> x+y+z == 2

        Returns:
            A string.
        """
        eqname = self.get_name()
        name_part = '{0}: '.format(eqname) if eqname else ''
        eq_var = self._binary_var
        if eq_var.is_generated():
            varname = 'x%d' % eq_var._index
        else:
            varname = str(eq_var)
        s_active_value = '' if self._active_value else '=0'
        s_symbol = self._symbol()
        return "{0}{1}{2} {4} [{3!s}]".format(name_part, varname, s_active_value, self.linear_constraint, s_symbol)


class IndicatorConstraint(LogicalConstraint):
    """ This class models indicator constraints.

    An indicator constraint links (one-way) the value of a binary variable to the satisfaction of a linear constraint.
    If the binary variable equals the active value, then the constraint is satisfied, but otherwise the constraint
    may or may not be satisfied.

    This class is not meant to be instantiated by the user.

    To create an indicator constraint, use the factory method :func:`docplex.mp.model.Model.add_indicator`
    defined on :class:`docplex.mp.model.Model`.

    """
    __slots__ = ()

    def __init__(self, model, binary_var, linear_ct, active_value=1, name=None):
        LogicalConstraint.__init__(self, model, binary_var, linear_ct, active_value, name)

    @property
    def short_typename(self):
        return "indicator constraint"

    def is_equivalence(self):
        return False

    def is_logical(self):
        return True

    def invalidate(self):
        """
        Sets the binary variable to the opposite of its active value.
        Typically used by indicator constraints with a trivial infeasible linear part.
        For example, z=1 -> 4 <= 3 sets z to 0 and
        z=0 -> 4 <= 3 sets z to 1.
        This is equivalent to if z=a => False, then z *cannot* be equal to a.
        """
        if 0 == self.active_value:
            # set to 1 : lb = 1
            self.binary_var.lb = 1
        elif 1 == self.active_value:
            # set to 0 ub = 0
            self.binary_var.ub = 0
        else:
            self.fatal("Unexpected active value for indicator constraint: {0!s}, value is: {1!s}, expecting 0 or 1",
                       # pragma: no cover
                       self, self.active_value)  # pragma: no cover

    def is_satisfied(self, solution, tolerance=1e-6):
        binary_value = solution.get_value(self._binary_var)
        if abs(binary_value - self._active_value) <= tolerance:
            is_ct_satisfied = self._linear_ct.is_satisfied(solution, tolerance)
            # only active if binary is active value:
            return abs(1 - is_ct_satisfied) <= tolerance
        else:
            # when binary var is not equal to active_value, indicator has no effect
            return True



class EquivalenceConstraint(LogicalConstraint):
    """ This class models equivalence constraints.

    An equivalence constraint links (both ways) the value of a binary variable
    to the satisfaction of a linear constraint.

    If the binary variable equals the truth value (default is 1),
    then the constraint is satisfied, conversely if the constraint
    is satisfied, the value of the variable is set to the truth value.

    This class is not meant to be instantiated by the user.

    """
    __slots__ = ()

    def __init__(self, model, binary_var, linear_ct, truth_value=1, name=None):
        LogicalConstraint.__init__(self, model, binary_var, linear_ct, truth_value, name)
        # connect exprs
        linear_ct.lock_discrete()

    @property
    def short_typename(self):
        return "equivalence constraint"

    def is_equivalence(self):
        return True

    def is_logical(self):
        return True

    def is_satisfied(self, solution, tolerance=1e-6):
        is_ct_satisfied = self._linear_ct.is_satisfied(solution, tolerance)
        binary_value = solution.get_value(self._binary_var)
        expected_value = self._active_value if is_ct_satisfied else 1 - self._active_value
        return ComparisonType.almost_equal(binary_value, expected_value, tolerance)


class IfThenConstraint(IndicatorConstraint):

    def __init__(self, model, if_ct, then_ct, negate=False):
        if_ct_status_var = if_ct.status_var
        self._if_ct = if_ct
        # if negated, then the then constraint is satisfied if not(if_ct) is satisfied
        # this is actually an if-else ...
        true_value = 0 if negate else 1
        IndicatorConstraint.__init__(self, model, binary_var=if_ct_status_var,
                                     linear_ct=then_ct, active_value=true_value, name=None)

    def to_string(self):
        return "{0!s} -> {1!s}".format(self._if_ct, self.linear_constraint)


class QuadraticConstraint(BinaryConstraint):
    """ The class models quadratic constraints.

        Quadratic constraints are of the form `<qexpr1> <OP> <qexpr2>`,
        where at least one of <qexpr1> or <qexpr2> is a quadratic expression.

    """

    def copy(self, target_model, var_map):
        # noinspection PyPep8
        copied_left_expr = self.left_expr.copy(target_model, var_map)
        copied_right_expr = self.right_expr.copy(target_model, var_map)
        copy_name = None if target_model.ignore_names else self.name
        return QuadraticConstraint(target_model, copied_left_expr, self.type, copied_right_expr, copy_name)

    def relaxed_copy(self, relaxed_model, var_map):
        raise DocplexLinearRelaxationError(self, cause='quadratic')

    def is_quadratic(self):
        return True

    @property
    def short_typename(self):
        return "quadratic constraint"

    def _get_index_scope(self):
        return self._model._quadct_scope

    __slots__ = ()

    def is_trivial(self):
        for qv, nqk in self.iter_net_quads():
            if nqk:
                return False
        # now check linear parts

        for lv, lk in self.iter_net_linear_coefs():
            if lk:
                return False
        return True

    def iter_net_linear_coefs(self):
        linear_left = self._left_expr.get_linear_part()
        linear_right = self._right_expr.get_linear_part()
        return self._iter_net_linear_coefs_sorted(linear_left, linear_right)

    def iter_net_quads(self):
        # INTERNAL
        left_expr = self._left_expr
        right_expr = self._right_expr
        if not right_expr.is_quad_expr():
            return left_expr.iter_sorted_quads()
        elif not left_expr.is_quad_expr():
            return right_expr.iter_opposite_ordered_quads()
        else:
            return self.generate_ordered_net_quads(left_expr, right_expr)

    def generate_ordered_net_quads(self, qleft, qright):
        # left first, then right
        for lqv, lqk in qleft.iter_sorted_quads():
            net_k = lqk - qright._get_quadratic_coefficient_from_var_pair(lqv)
            if 0 != net_k:
                yield lqv, net_k
        for rqv, rqk in qright.iter_sorted_quads():
            if not qleft.contains_quad(rqv) and 0 != rqk:
                yield rqv, -rqk

    def _set_left_expr(self, new_left_expr):
        self.get_quadratic_factory().set_quadratic_constraint_expr_from_pos(self, pos=0, new_expr=new_left_expr)

    left_expr = property(BinaryConstraint.get_left_expr, _set_left_expr)

    def _set_right_expr(self, new_right_expr):
        self.get_quadratic_factory().set_quadratic_constraint_expr_from_pos(self, pos=1, new_expr=new_right_expr)

    right_expr = property(BinaryConstraint.get_right_expr, _set_right_expr)

    @property
    def benders_annotation(self):
        """
        This property is used to get or set the Benders annotation of a constraint.
        The value of the annotation must be a positive integer

        """
        return self.get_benders_annotation()

    @benders_annotation.setter
    def benders_annotation(self, new_anno):
        self.set_benders_annotation(new_anno)

    @property
    def slack_value(self):
        """ This property returns the slack value of the constraint.

        Note:
            This method will raise an exception if the model has not been solved successfully.
        """
        return self._model._slack_value1(self)

    @property
    def sense(self):
        """ This property is used to get or set the sense of the constraint; sense is an enumerated value
        of type :class:`ComparisonType`, with three possible values:

        - LE for e1 <= e2 constraints

        - EQ for e1 == e2 constraints

        - GE for e1 >= e2 constraints

        where e1 and e2 denote quadratic expressions.

        """
        return self._ctsense

    @sense.setter
    def sense(self, new_sense):
        self.set_sense(new_sense)

    def set_sense(self, new_sense):
        self.get_quadratic_factory().set_quadratic_constraint_sense(self, new_sense)

    # compat
    @property
    def type(self):
        return self._ctsense

    def has_net_quadratic_term(self):
        # INTERNAL
        return any(nk for _, nk in self.iter_net_quads())

    def notify_expr_modified(self, expr, event):
        # INTERNAL
        self.get_quadratic_factory().update_quadratic_constraint(self, expr, event)

    def notify_expr_replaced(self, old_expr, new_expr):
        qfact = self.get_quadratic_factory()
        if old_expr is self._left_expr:
            qfact.set_quadratic_constraint_expr_from_pos(qct=self, pos=0, new_expr=new_expr,
                                                                          update_subscribers=False)
        elif old_expr is self._right_expr:
            qfact.set_quadratic_constraint_expr_from_pos(qct=self, pos=1, new_expr=new_expr,
                                                                          update_subscribers=False)
        else:
            # should not happen
            pass
        # new expr takes al subscribers from old expr
        new_expr.grab_subscribers(old_expr)


class PwlConstraint(AbstractConstraint):
    """ This class models piecewise linear constraints.

    This class is not meant to be instantiated by the user.
    To create a piecewise constraint, use the factory method :func:`docplex.mp.model.Model.piecewise`
    defined on :class:`docplex.mp.model.Model`.

    """

    __slots__ = ('_pwl_expr', '_input_var', '_y')

    def __init__(self, model, pwl_expr, name=None):
        AbstractConstraint.__init__(self, model, name)
        self._pwl_expr = pwl_expr
        self._input_var = pwl_expr._x_var
        self._y = None

    def resolve(self):
        self._pwl_expr.resolve()

    def is_satisfied(self, solution, tolerance):
        expr_value = self._input_var._get_solution_value(solution)
        y_value = solution._get_var_value(self._y)
        computed_f_expr_value = self.pwl_func.evaluate(expr_value)
        return ComparisonType.almost_equal(y_value, computed_f_expr_value, tolerance)

    @property
    def expr(self):
        """ This property returns the linear expression of the piecewise linear constraint.
        """
        return self._input_var

    @property
    def pwl_func(self):
        """ This property returns the piecewise linear function of the piecewise linear constraint.
        """
        return self._pwl_expr.pwl_func

    @property
    def y(self):
        """ This property returns the output variable associated with the piecewise linear constraint.
        """
        if self._y is None:
            self._y = self._pwl_expr._get_allocated_f_var()
        return self._y

    @property
    def usage_counter(self):
        return self._pwl_expr.usage_counter

    def _get_index_scope(self):
        return self._model._pwl_scope

    def iter_exprs(self):
        yield self._pwl_expr

    def iter_variables(self):
        """Iterates over all the variables of the piecewise linear constraint.

        Returns:
           An iterator object.
        """
        yield self.y
        yield self._input_var


    def iter_extended_variables(self):
        # iterates on all extended variables involved in the computation
        # yvar, xavr, plus all argument rexpr vars
        # if argument var is identical to the input var, it is returned twice...?
        yield self.y
        yield self._input_var
        for v in self._pwl_expr._argument_expr.iter_variables():
            yield v


    def get_var_coef(self, dvar):
        if dvar is self.y:
            return 1
        else:
            return self.expr.unchecked_get_coef(dvar)

    def copy(self, target_model, var_map):
        # Internal: copy must not be invoked on PwlConstraint.
        raise NotImplementedError  # pragma: no cover

    def relaxed_copy(self, relaxed_model, var_map):
        raise DocplexLinearRelaxationError(self, cause='pwl')

    def to_string(self):
        pwlf = self.pwl_func
        pwlf_s = pwlf.name or 'pwl?'
        return "{0} == {1!s}({2!s})".format(self.y, pwlf_s, self.expr)

    def __str__(self):
        """ Returns a string representation of the piecewise linear constraint.

        Example:
            `y == pwl_name(x + z)` represents the piecewise linear constraint where the variable `y` is
            constrained to be equal to the value of the piecewise linear function whose name is 'pwl_name'
            applied to the expression (x + z).

        Returns:
            A string.
        """
        return self.to_string()

    def __repr__(self):
        printable_name = self.safe_name
        return "docplex.mp.PwlConstraint[{0}]({1},{2!s},{3})".\
            format(printable_name, self.y, self.pwl_func, self.expr)
